from ._arcgis_model import ArcGISModel, _get_device
from pathlib import Path
import json
from ._codetemplate import code
import warnings
from .._data import _raise_fastai_import_error
import traceback

import logging

logger = logging.getLogger()

HAS_OPENCV = True
HAS_FASTAI = True

try:
    import torch
    from torch import tensor, Tensor
    import numpy as np
    from fastai.vision.learner import cnn_learner
    from fastai.callbacks.hooks import model_sizes
    from fastai.vision.learner import create_body, cnn_config
    from fastai.vision import ImageList
    from fastai.vision import imagenet_stats
    from fastai.vision.image import bb2hw, Image, pil2tensor
    from torchvision import models
    from .._utils.pascal_voc_rectangles import (
        ObjectDetectionCategoryList,
        show_results_multispectral,
    )
    from ._ssd_utils import (
        SSDHead,
        BCE_Loss,
        FocalLoss,
        postprocess,
    )
    from ._ssd_utils import (
        compute_class_AP,
        SSDHeadv2,
        kmeans,
        avg_iou,
        AveragePrecision,
    )
    from ._arcgis_model import (
        _set_multigpu_callback,
        _resnet_family,
        _vgg_family,
        _densenet_family,
    )
    from ._timm_utils import timm_config, filter_timm_models, _get_feature_size
    from torch.nn import Module as NnModule
    import PIL
    from .._image_utils import (
        _get_image_chips,
        _get_transformed_predictions,
        _draw_predictions,
        _exclude_detection,
    )
    from .._video_utils import VideoUtils
    from .._utils.common import (
        get_multispectral_data_params_from_emd,
        _get_emd_path,
        read_image,
    )
    from ._inferencing.util import actn_to_bb, hw2corners, nms_jit
    from fastprogress.fastprogress import progress_bar
    from .._utils.env import is_arcgispronotebook
    import matplotlib.pyplot as plt
    from .._utils.utils import chips_to_batch
    from .._utils.pascal_voc_rectangles import _reconstruct
except Exception as e:
    import_exception = "\n".join(
        traceback.format_exception(type(e), e, e.__traceback__)
    )

    class NnModule:
        pass

    HAS_FASTAI = False

try:
    import cv2
except Exception:
    HAS_OPENCV = False


def _mobilenet_split(m: NnModule):
    return m[0][0][0], m[1]


from typing import Union, Tuple, Optional, List


try:

    @torch.jit.script
    def _nms_jit_loop(
        conf_scores, a_ic, nms_overlap: float, thres: float
    ) -> Tuple[List[Tensor], List[Tensor], List[List[Tensor]]]:
        out1, out2 = [], []
        cc = torch.jit.annotate(List[List[Tensor]], [])
        for cl in range(1, len(conf_scores)):
            c_mask = conf_scores[cl] > thres
            if c_mask.sum() == 0:
                continue
            scores = conf_scores[cl][c_mask]
            l_mask = c_mask.unsqueeze(1).expand_as(a_ic)
            boxes = a_ic[l_mask].view(-1, 4)

            ids, count = nms_jit(
                boxes.data, scores, nms_overlap, 50
            )  # FIX- NMS overlap hardcoded #TODO: jit
            ids = ids[: int(count.item())]
            out1.append(scores[ids])
            out2.append(boxes.data[ids])
            cc.append(
                [
                    torch.tensor([cl]).to(conf_scores.device).int()
                    for i in range(int(count.item()))
                ]
            )

        return out1, out2, cc

    @torch.jit.script
    def _get_jit_predictions(
        conf_scores, a_ic, nms_overlap: float, thres: float
    ) -> Tuple[Tensor, Tensor, Tensor]:
        out1, out2, cc = _nms_jit_loop(conf_scores, a_ic, nms_overlap, thres)
        if len(cc) == 0:
            cc = [[torch.tensor([0]).int().to(a_ic.device)]]
        cc_temp = torch.jit.annotate(List[Tensor], [])
        for c in range(len(cc)):
            cc_temp.append(torch.cat(cc[c]))
        cc_concat = torch.cat(cc_temp)
        if len(out1) == 0:
            out1 = [torch.empty((0,), dtype=conf_scores.dtype).to(conf_scores.device)]
        out1_concat = torch.cat(out1)
        if len(out2) == 0:
            out2 = [torch.empty((0,), dtype=conf_scores.dtype).to(conf_scores.device)]
        out2_concat = torch.cat(out2)
        bbox, clas, prs, thresh = (
            out2_concat,
            cc_concat,
            out1_concat,
            thres,
        )  # FIX- hardcoded threshold
        return bbox, clas, prs

    @torch.jit.script
    def _process_jit_bboxes(bboxes: List[Tensor]):
        # print(bboxes)
        processed_bboxes = []
        for batch in range(len(bboxes)):
            for bbox in bboxes[batch]:
                output = torch.clone(bbox)
                bbox[0] = output[1]
                bbox[1] = output[0]
                bbox[2] = output[3] - output[1]
                bbox[3] = output[2] - output[0]
                processed_bboxes.append(bbox)
        if not len(processed_bboxes) == 0:
            out_bboxes = torch.stack([bbox for bbox in processed_bboxes])
            return out_bboxes
        else:
            return torch.empty((len(bboxes), 0, 0, 0)).float()

except:
    print("torch not available\n")


class SSDTracer(torch.nn.Module):
    def __init__(self, model, device, chip_size, anchors, grid_sizes, trace=False):
        super().__init__()
        if trace:
            self.model = torch.jit.trace(
                model.to(device),
                torch.rand(1, 3, chip_size[0], chip_size[1]).to(device),
            )
        else:
            self.model = model.to(device)
        self.anchors = anchors.to(device)
        self.grid_sizes = grid_sizes.to(device)

    def _get_nms_preds(
        self,
        b_clas,
        b_bb,
        idx: int,
        anchors,
        grid_sizes,
        nms_overlap: float,
        thres: float,
    ):
        a_ic = actn_to_bb(b_bb[idx], anchors, grid_sizes)
        conf_scores = b_clas[idx].sigmoid().t().detach().to(anchors.device)
        out = _get_jit_predictions(conf_scores, a_ic, nms_overlap, thres)
        return out

    # TODO: batch_size
    def process_bboxes(self, batch_output: List[Tensor]) -> List[Tensor]:
        num_boxes = 0
        pred_bboxes = []
        pred_labels = []
        pred_scores = []

        batch = 0

        for chip_idx, output in enumerate(batch_output[1]):
            pred_bbox, pred_label, pred_score = self._get_nms_preds(
                batch_output[0],
                batch_output[1],
                chip_idx,
                self.anchors,
                self.grid_sizes,
                nms_overlap=0.1,
                thres=0.3,
            )

            batch += 1
            pred_bboxes.append(pred_bbox)
            pred_labels.append(pred_label)
            pred_scores.append(pred_score)

        if not len(pred_bboxes) == 0:
            pred_bboxes_final = _process_jit_bboxes(pred_bboxes).to(
                batch_output[0].device
            )
            pred_labels_final = torch.stack([label for label in pred_labels]).to(
                batch_output[0].device
            )
            pred_scores_final = torch.stack([score for score in pred_scores]).to(
                batch_output[0].device
            )  # TODO
            return pred_bboxes_final, pred_labels_final, pred_scores_final
        else:
            dummy = torch.empty((batch, 0, 0, 0)).float().to(batch_output[0].device)
            return dummy, dummy, dummy

    def forward(self, inp: Tensor) -> List[Tensor]:
        out = self.model(inp)
        out_final = self.process_bboxes(out)
        return out_final


class SingleShotDetector(ArcGISModel):

    """
    Creates a Single Shot Detector with the specified grid sizes, zoom scales
    and aspect ratios. Based on Fast.ai MOOC Version2 Lesson 9.

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    data                    Required fastai Databunch. Returned data object from
                            :meth:`~arcgis.learn.prepare_data` function.
    ---------------------   -------------------------------------------
    grids                   Required list. Grid sizes used for creating anchor
                            boxes.
    ---------------------   -------------------------------------------
    zooms                   Optional list. Zooms of anchor boxes.
    ---------------------   -------------------------------------------
    ratios                  Optional list of tuples. Aspect ratios of anchor
                            boxes.
    ---------------------   -------------------------------------------
    backbone                Optional string. Backbone convolutional neural network
                            model used for feature extraction, which
                            is `resnet34` by default.
                            Supported backbones: ResNet, DenseNet, VGG families
                            and specified Timm models(experimental support) from
                            :func:`~arcgis.learn.SingleShotDetector.backbones`.
    ---------------------   -------------------------------------------
    dropout                 Optional float. Dropout probability. Increase it to
                            reduce overfitting.
    ---------------------   -------------------------------------------
    bias                    Optional float. Bias for SSD head.
    ---------------------   -------------------------------------------
    focal_loss              Optional boolean. Uses Focal Loss if True.
    ---------------------   -------------------------------------------
    pretrained_path         Optional string. Path where pre-trained model is
                            saved.
    ---------------------   -------------------------------------------
    location_loss_factor    Optional float. Sets the weight of the bounding box
                            loss. This should be strictly between 0 and 1. This
                            is default `None` which gives equal weight to both
                            location and classification loss. This factor
                            adjusts the focus of model on the location of
                            bounding box.
    ---------------------   -------------------------------------------
    ssd_version             Optional int within [1,2]. Use version=1 for arcgis v1.6.2 or earlier
    ---------------------   -------------------------------------------
    backend                 Optional string. Controls the backend framework to be used
                            for this model, which is 'pytorch' by default.

                            valid options are 'pytorch', 'tensorflow'
    =====================   ===========================================

    :return:
        :class:`~arcgis.learn.SingleShotDetector` Object
    """

    def __init__(
        self,
        data,
        grids=None,
        zooms=[1.0],
        ratios=[[1.0, 1.0]],
        backbone=None,
        drop=0.3,
        bias=-4.0,
        focal_loss=False,
        pretrained_path=None,
        location_loss_factor=None,
        ssd_version=2,
        backend="pytorch",
        *args,
        **kwargs,
    ):
        super().__init__(data, backbone, pretrained_path=pretrained_path, **kwargs)
        data = self._data

        if pretrained_path is not None:
            backbone_pretrained = False
        else:
            backbone_pretrained = True

        self._backend = backend
        if self._backend == "tensorflow":
            self._intialize_tensorflow(
                data,
                grids,
                zooms,
                ratios,
                backbone,
                drop,
                bias,
                pretrained_path,
                location_loss_factor,
            )
        else:
            self._check_dataset_support(self._data)
            if not (self._check_backbone_support(getattr(self, "_backbone", backbone))):
                raise Exception(
                    f"Enter only compatible backbones from {', '.join(self.supported_backbones)}"
                )

            # assert (location_loss_factor is not None) or ((location_loss_factor > 0) and (location_loss_factor < 1)),
            if not ssd_version in [1, 2]:
                raise Exception("ssd_version can be only [1,2]")

            if location_loss_factor is not None:
                if not ((location_loss_factor > 0) and (location_loss_factor < 1)):
                    raise Exception(
                        "`location_loss_factor` should be greater than 0 and less than 1"
                    )
            self.location_loss_factor = location_loss_factor

            self._code = code
            self.ssd_version = ssd_version

            if "timm" in self._backbone.__module__:
                timm_meta = timm_config(self._backbone)
                backbone_cut = timm_meta["cut"]
                backbone_split = timm_meta["split"]
            else:
                backbone_cut = None
                backbone_split = None

            backbone_name = self._backbone.__name__[:3]

            if self._backbone.__name__ == "mobilenet_v2":
                backbone_cut = -1
                backbone_split = _mobilenet_split

            if ssd_version == 1:
                if grids == None:
                    grids = [4, 2, 1]

                self._create_anchors(grids, zooms, ratios)

                feature_sizes = model_sizes(
                    create_body(self._backbone, cut=backbone_cut, pretrained=False),
                    size=(data.chip_size, data.chip_size),
                )
                num_features = feature_sizes[-1][-1]
                num_channels = feature_sizes[-1][1]

                ssd_head = SSDHead(
                    grids,
                    self._anchors_per_cell,
                    data.c,
                    num_features=num_features,
                    drop=drop,
                    bias=bias,
                    num_channels=num_channels,
                )
            elif ssd_version == 2:
                # find bounding boxes height and width

                if grids is None:
                    logger.info("Computing optimal grid size...")
                    hw = data.height_width
                    hw = np.array(hw)

                    # find most suitable centroids for dataset
                    centroid = kmeans(hw, 1)
                    avg = avg_iou(hw, centroid)

                    for num_anchor in range(2, 5):
                        new_centroid = kmeans(hw, num_anchor)
                        new_avg = avg_iou(hw, new_centroid)
                        if (new_avg - avg) < 0.05:
                            break
                        avg = new_avg
                        centroid = new_centroid.copy()

                    centroid = np.sort(np.max(centroid, axis=1))
                    centroid = centroid[centroid != 0]

                    # find grid size

                    grids = list(
                        map(
                            int,
                            map(
                                round,
                                data.chip_size / centroid,
                            ),
                        )
                    )
                    grids = list(set(grids))
                    grids.sort(reverse=True)
                    if grids[-1] == 0:
                        grids[-1] = 1
                    grids = list(set(grids))

                self._create_anchors(grids, zooms, ratios)

                feature_sizes = _get_feature_size(
                    self._orig_backbone
                    if hasattr(self, "_orig_backbone")
                    else self._backbone,
                    cut=backbone_cut,
                    chip_size=(data.chip_size, data.chip_size),
                )

                num_features = feature_sizes[-1][-1]
                num_channels = feature_sizes[-1][1]

                if (
                    grids[0] > 8
                    and abs(num_features - grids[0]) > 4
                    and backbone_name == "res"
                    and "bit" not in self._backbone.__name__
                ):
                    num_features = feature_sizes[-2][-1]
                    num_channels = feature_sizes[-2][1]
                    backbone_cut = -3
                ssd_head = SSDHeadv2(
                    grids,
                    self._anchors_per_cell,
                    data.c,
                    num_features=num_features,
                    drop=drop,
                    bias=bias,
                    num_channels=num_channels,
                )

            else:
                raise Exception("SSDVersion can only be 1 or 2")

            if hasattr(self, "_backbone_ms"):
                self._orig_backbone = self._backbone
                self._backbone = self._backbone_ms

            if (
                hasattr(self, "_orig_backbone")
                and "densenet" in self._orig_backbone.__name__
            ):
                backbone_cut = cnn_config(self._orig_backbone)["cut"]
                backbone_split = cnn_config(self._orig_backbone)["split"]

            self.learn = cnn_learner(
                data=data,
                base_arch=self._backbone,
                cut=backbone_cut,
                pretrained=backbone_pretrained,
                split_on=backbone_split,
                custom_head=ssd_head,
            )
            self.learn.metrics = [AveragePrecision(self, data.c - 1)]
            self._arcgis_init_callback()  # make first conv weights learnable

            if focal_loss:
                self._loss_f = FocalLoss(data.c)
            else:
                self._loss_f = BCE_Loss(data.c)
            self.learn.loss_func = self._ssd_loss

            _set_multigpu_callback(self)
            if pretrained_path is not None:
                self.load(pretrained_path)

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return "<%s>" % (type(self).__name__)

    @property
    def supported_backbones(self):
        """Supported list of backbones for this model."""
        return SingleShotDetector._supported_backbones()

    @staticmethod
    def backbones():
        """Supported list of backbones for this model."""
        return SingleShotDetector._supported_backbones()

    @staticmethod
    def _supported_backbones():
        timm_models = filter_timm_models(["*repvgg*", "*tresnet*"])
        timm_backbones = list(map(lambda m: "timm:" + m, timm_models))

        return [
            *_resnet_family,
            *_densenet_family,
            *_vgg_family,
            models.mobilenet_v2.__name__,
        ] + timm_backbones

    @property
    def supported_datasets(self):
        """Supported dataset types for this model."""
        return SingleShotDetector._supported_datasets()

    @staticmethod
    def _supported_datasets():
        return ["PASCAL_VOC_rectangles", "KITTI_rectangles"]

    @staticmethod
    def _available_metrics():
        return ["valid_loss", "average_precision"]

    @classmethod
    def from_model(cls, emd_path, data=None):
        """
        Creates a Single Shot Detector from an Esri Model Definition (EMD) file.

        Note: Only supported for Pytorch models.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to Deep Learning Package
                                (DLPK) or Esri Model Definition(EMD) file.
        ---------------------   -------------------------------------------
        data                    Required fastai Databunch or None. Returned data
                                object from :meth:`~arcgis.learn.prepare_data` function or None for
                                inferencing.
        =====================   ===========================================

        :return:
            :class:`~arcgis.learn.SingleShotDetector` Object
        """
        return cls.from_emd(data, emd_path)

    @classmethod
    def from_emd(cls, data, emd_path):
        """
        Creates a Single Shot Detector from an Esri Model Definition (EMD) file.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        data                    Required fastai Databunch or None. Returned data
                                object from :meth:`~arcgis.learn.prepare_data` function or None for
                                inferencing.
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to Esri Model Definition
                                file.
        =====================   ===========================================

        :return: :class:`~arcgis.learn.SingleShotDetector` Object
        """
        if not HAS_FASTAI:
            _raise_fastai_import_error(import_exception=import_exception)

        emd_path = _get_emd_path(emd_path)
        emd = json.load(open(emd_path))
        model_file = Path(emd["ModelFile"])
        backbone = emd.get("backbone", "resnet34")
        ssd_version = int(emd.get("SSDVersion", 1))
        chip_size = emd["ImageWidth"]

        if not model_file.is_absolute():
            model_file = emd_path.parent / model_file

        class_mapping = {i["Value"]: i["Name"] for i in emd["Classes"]}

        resize_to = emd.get("resize_to")
        if isinstance(resize_to, list):
            resize_to = (resize_to[0], resize_to[1])

        # Tensorflow support
        backend = emd.get("ModelParameters", {}).get("backend", "pytorch")
        if backend == "tensorflow":
            backbone = emd["ModelParameters"].get("backbone", "ResNet50")

        data_passed = True
        # Create an image Staunch for when loading the model using emd (without training data)
        if data is None:
            data_passed = False
            train_tfms = []
            val_tfms = []
            ds_tfms = (train_tfms, val_tfms)

            with warnings.catch_warnings():
                warnings.simplefilter("ignore", UserWarning)

                sd = ImageList([], path=emd_path.parent.parent.parent).split_by_idx([])
                data = (
                    sd.label_const(
                        0,
                        label_cls=ObjectDetectionCategoryList,
                        classes=list(class_mapping.values()),
                    )
                    .transform(ds_tfms)
                    .databunch(device=_get_device())
                    .normalize(imagenet_stats)
                )

            data.chip_size = chip_size
            data.class_mapping = class_mapping
            data.classes = ["background"] + list(class_mapping.values())
            data._is_empty = True
            # Add 1 for background class
            data.c += 1
            data.emd_path = emd_path
            data.emd = emd
            data = get_multispectral_data_params_from_emd(data, emd)

        data.resize_to = resize_to

        ssd = cls(
            data,
            emd["Grids"],
            emd["Zooms"],
            emd["Ratios"],
            pretrained_path=str(model_file),
            backend=backend,
            backbone=backbone,
            ssd_version=ssd_version,
        )

        if not data_passed:
            ssd.learn.data.single_ds.classes = ssd._data.classes
            ssd.learn.data.single_ds.y.classes = ssd._data.classes

        return ssd

    def _create_anchors(self, anc_grids, anc_zooms, anc_ratios):
        self.grids = anc_grids
        self.zooms = anc_zooms
        self.ratios = anc_ratios

        anchor_scales = [
            (anz * i, anz * j) for anz in anc_zooms for (i, j) in anc_ratios
        ]

        self._anchors_per_cell = len(anchor_scales)

        anc_offsets = [1 / (o * 2) for o in anc_grids]

        anc_x = np.concatenate(
            [
                np.repeat(np.linspace(ao, 1 - ao, ag), ag)
                for ao, ag in zip(anc_offsets, anc_grids)
            ]
        )
        anc_y = np.concatenate(
            [
                np.tile(np.linspace(ao, 1 - ao, ag), ag)
                for ao, ag in zip(anc_offsets, anc_grids)
            ]
        )
        anc_ctrs = np.repeat(
            np.stack([anc_x, anc_y], axis=1), self._anchors_per_cell, axis=0
        )

        anc_sizes = np.concatenate(
            [
                np.array(
                    [[o / ag, p / ag] for i in range(ag * ag) for o, p in anchor_scales]
                )
                for ag in anc_grids
            ]
        )

        self._grid_sizes = (
            torch.Tensor(
                np.concatenate(
                    [
                        np.array(
                            [1 / ag for i in range(ag * ag) for o, p in anchor_scales]
                        )
                        for ag in anc_grids
                    ]
                )
            )
            .unsqueeze(1)
            .to(self._device)
        )

        self._anchors = (
            torch.Tensor(np.concatenate([anc_ctrs, anc_sizes], axis=1))
            .float()
            .to(self._device)
        )

        self._anchor_cnr = self._hw2corners(self._anchors[:, :2], self._anchors[:, 2:])

    def _hw2corners(self, ctr, hw):
        return torch.cat([ctr - hw / 2, ctr + hw / 2], dim=1)

    def _get_y(self, bbox, clas):
        try:
            bbox = bbox.view(-1, 4)  # /sz
        except Exception:
            bbox = torch.zeros(size=[0, 4])
        bb_keep = ((bbox[:, 2] - bbox[:, 0]) > 0).nonzero()[:, 0]
        return bbox[bb_keep], clas[bb_keep]

    def _actn_to_bb(self, actn, anchors, grid_sizes):
        actn_bbs = torch.tanh(actn)
        actn_centers = (actn_bbs[:, :2] / 2 * grid_sizes) + anchors[:, :2]
        actn_hw = (actn_bbs[:, 2:] / 2 + 1) * anchors[:, 2:]
        return self._hw2corners(actn_centers, actn_hw)

    def _map_to_ground_truth(self, overlaps, print_it=False):
        prior_overlap, prior_idx = overlaps.max(1)
        if print_it:
            print(prior_overlap)
        gt_overlap, gt_idx = overlaps.max(0)
        gt_overlap[prior_idx] = 1.99
        for i, o in enumerate(prior_idx):
            gt_idx[o] = i
        return gt_overlap, gt_idx

    def _ssd_1_loss(self, b_c, b_bb, bbox, clas, print_it=False):
        bbox, clas = self._get_y(bbox, clas)
        bbox = self._normalize_bbox(bbox)

        a_ic = self._actn_to_bb(b_bb, self._anchors, self._grid_sizes)
        overlaps = self._jaccard(bbox.data, self._anchor_cnr.data)
        try:
            gt_overlap, gt_idx = self._map_to_ground_truth(overlaps, print_it)
        except Exception as e:
            return (
                torch.tensor(0.0, requires_grad=True).to(self._device),
                torch.tensor(0.0, requires_grad=True).to(self._device),
            )
        gt_clas = clas[gt_idx]
        pos = gt_overlap > 0.4
        pos_idx = torch.nonzero(pos)[:, 0]
        gt_clas[~pos] = 0
        gt_bbox = bbox[gt_idx]
        loc_loss = ((a_ic[pos_idx] - gt_bbox[pos_idx]).abs()).mean()
        clas_loss = self._loss_f(b_c, gt_clas)
        return loc_loss, clas_loss

    def _ssd_loss(self, pred, targ1, targ2, print_it=False):
        lcs, lls = 0.0, 0.0
        for b_c, b_bb, bbox, clas in zip(*pred, targ1, targ2):
            loc_loss, clas_loss = self._ssd_1_loss(
                b_c, b_bb, bbox.to(self._device), clas.to(self._device), print_it
            )
            lls += loc_loss
            lcs += clas_loss
        if print_it:
            print("loc: {lls}, clas: {lcs}".format(lls=lls, lcs=lcs))
        if self.location_loss_factor is None:
            return lls + lcs
        else:
            return (
                self.location_loss_factor * lls + (1 - self.location_loss_factor) * lcs
            )

    def _intersect(self, box_a, box_b):
        max_xy = torch.min(box_a[:, None, 2:], box_b[None, :, 2:])
        min_xy = torch.max(box_a[:, None, :2], box_b[None, :, :2])
        inter = torch.clamp((max_xy - min_xy), min=0)
        return inter[:, :, 0] * inter[:, :, 1]

    def _box_sz(self, b):
        return (b[:, 2] - b[:, 0]) * (b[:, 3] - b[:, 1])

    def _jaccard(self, box_a, box_b):
        inter = self._intersect(box_a, box_b)
        union = (
            self._box_sz(box_a).unsqueeze(1) + self._box_sz(box_b).unsqueeze(0) - inter
        )
        return inter / union

    def _normalize_bbox(self, bbox):
        return (bbox + 1.0) / 2.0

    @property
    def _model_metrics(self):
        return {
            "average_precision_score": self.average_precision_score(show_progress=True)
        }

    def _analyze_pred(
        self, pred, thresh=0.5, nms_overlap=0.1, ret_scores=True, device=None
    ):
        return postprocess(
            pred,
            model=self,
            thresh=thresh,
            nms_overlap=nms_overlap,
            ret_scores=ret_scores,
            device=device,
        )

    def _save_device_model(self, model, device, save_path):
        model.eval()
        if hasattr(self._data, "chip_size"):
            chip_size = self._data.chip_size
            if not isinstance(chip_size, tuple):
                chip_size = (chip_size, chip_size)
        inp = torch.rand(1, 3, chip_size[0], chip_size[1]).to(device)
        model = SSDTracer(model, device, chip_size, self._anchors, self._grid_sizes)
        model = model.to(device)
        traced_model = None
        with torch.no_grad():
            traced_model = self._trace(model, inp, True)
        torch.jit.save(traced_model, save_path)

        return traced_model

    def _save_pytorch_torchscript(self, name, save=True):
        traced_model_cpu = None
        traced_model_gpu = None

        model = self.learn.model
        model.eval()
        device = self._device

        cpu = torch.device("cpu")
        save_path_cpu = (
            self.learn.path / self.learn.model_dir / f"{name}-cpu.pt"
        ).__str__()
        traced_model_cpu = self._save_device_model(model, cpu, save_path_cpu)
        save_path_cpu = f"{name}-cpu.pt"

        save_path_gpu = ""
        if torch.cuda.is_available():
            gpu = torch.device("cuda")
            save_path_gpu = (
                self.learn.path / self.learn.model_dir / f"{name}-gpu.pt"
            ).__str__()
            traced_model_gpu = self._save_device_model(model, gpu, save_path_gpu)
            save_path_gpu = f"{name}-gpu.pt"

        model.to(device)

        if not save:
            return [traced_model_cpu, traced_model_gpu]
        return [save_path_cpu, save_path_gpu]

    def _save_pytorch_tflite(self, name):
        import tensorflow as tf
        import logging

        tf.get_logger().setLevel(logging.ERROR)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            import onnx
            import onnx_tf
            from onnx_tf.backend import prepare

        traced_models = self._save_pytorch_torchscript(name, False)
        if traced_models[0] is None:
            return ["", ""]

        cpu = torch.device("cpu")
        device = cpu
        traced_model = traced_models[0].to(device)
        if hasattr(self._data, "chip_size"):
            chip_size = self._data.chip_size
            if not isinstance(chip_size, tuple):
                chip_size = (chip_size, chip_size)
        num_input_channels = list(self.learn.model.parameters())[0].shape[1]
        inp = torch.randn([1, num_input_channels, chip_size[0], chip_size[1]]).to(
            device
        )

        save_path_tflite = self.learn.path / self.learn.model_dir / f"{name}.tflite"
        save_path_onnx = self.learn.path / self.learn.model_dir / f"{name}.onnx"
        save_path_pb = self.learn.path / self.learn.model_dir / f"{name}"

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            torch.onnx.export(
                traced_model,
                inp,
                save_path_onnx,
                export_params=True,
                do_constant_folding=False,
                verbose=True,
                input_names=["input"],
                output_names=["output"],
                opset_version=11,
            )

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            arcgis_onnx = onnx.load(save_path_onnx)
            tf_onnx = prepare(arcgis_onnx, logging_level="ERROR")
            tf_onnx.export_graph(str(save_path_pb))

        converter = tf.lite.TFLiteConverter.from_saved_model(str(save_path_pb))
        converter.experimental_new_converter = True
        converter.optimizations = [tf.compat.v1.lite.Optimize.DEFAULT]
        converter.target_ops = [
            tf.lite.OpsSet.TFLITE_BUILTINS_INT8,
            tf.lite.OpsSet.SELECT_TF_OPS,
        ]
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            tflite_model = converter.convert()
        with tf.io.gfile.GFile(save_path_tflite, "wb") as f:
            f.write(tflite_model)

        return [save_path_tflite, save_path_onnx]

    def _get_emd_params(self, save_inference_file):
        import random

        _emd_template = {}
        _emd_template["Framework"] = "arcgis.learn.models._inferencing"
        if save_inference_file:
            _emd_template["InferenceFunction"] = "ArcGISObjectDetector.py"
        else:
            _emd_template[
                "InferenceFunction"
            ] = "[Functions]System\\DeepLearning\\ArcGISLearn\\ArcGISObjectDetector.py"
        _emd_template["ModelConfiguration"] = "_DynamicSSD"
        _emd_template["ModelType"] = "ObjectDetection"
        _emd_template["ExtractBands"] = [0, 1, 2]
        if "timm" in self._backbone.__module__:
            bckbn_name = "timm:" + self._backbone.__name__
        else:
            bckbn_name = self._backbone.__name__
        _emd_template["backbone"] = bckbn_name
        if _emd_template["backbone"] == "backbone_wrapper":
            _emd_template["backbone"] = self._orig_backbone.__name__
        _emd_template["Grids"] = self.grids
        _emd_template["Zooms"] = self.zooms
        _emd_template["Ratios"] = self.ratios
        _emd_template["SSDVersion"] = self.ssd_version
        _emd_template["Classes"] = []

        class_data = {}
        for i, class_name in enumerate(
            self._data.classes[1:]
        ):  # 0th index is background
            inverse_class_mapping = {v: k for k, v in self._data.class_mapping.items()}
            class_data["Value"] = inverse_class_mapping[class_name]
            class_data["Name"] = class_name
            color = [random.choice(range(256)) for i in range(3)]
            class_data["Color"] = color
            _emd_template["Classes"].append(class_data.copy())

        return _emd_template

    def _get_tfonnx_emd_params(self):
        return {"ModelConfiguration": "_SSDTensorflow"}

    def show_results(self, rows=5, thresh=0.5, nms_overlap=0.1):
        """
        Displays the results of a trained model on a part of the validation set.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        rows                    Optional int. Number of rows of results
                                to be displayed.
        ---------------------   -------------------------------------------
        thresh                  Optional float. The probability above which
                                a detection will be considered valid.
        ---------------------   -------------------------------------------
        nms_overlap             Optional float. The intersection over union
                                threshold with other predicted bounding
                                boxes, above which the box with the highest
                                score will be considered a true positive.
        =====================   ===========================================

        """
        self._check_requisites()
        if rows > len(self._data.valid_ds):
            rows = len(self._data.valid_ds)
        self.learn.show_results(
            rows=rows, thresh=thresh, nms_overlap=nms_overlap, model=self
        )
        if is_arcgispronotebook():
            plt.show()

    def _show_results_multispectral(
        self, rows=5, thresh=0.3, nms_overlap=0.1, alpha=1, **kwargs
    ):
        """
        Displays the results of a trained model on a part of the validation set.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        rows                    Optional int. Number of rows of results
                                to be displayed.
        ---------------------   -------------------------------------------
        thresh                  Optional float. The probability above which
                                a detection will be considered valid.
        ---------------------   -------------------------------------------
        nms_overlap             Optional float. The intersection over union
                                threshold with other predicted bounding
                                boxes, above which the box with the highest
                                score will be considered a true positive.
        ---------------------   -------------------------------------------
        alpha                   Optional Float.
                                Opacity of the lables for the corresponding
                                images. Values range between 0 and 1, where
                                1 means opaque.
        =====================   ===========================================

        """
        return_fig = kwargs.get("return_fig", False)
        ret_val = show_results_multispectral(
            self,
            nrows=rows,
            thresh=thresh,
            nms_overlap=nms_overlap,
            alpha=alpha,
            **kwargs,
        )
        if return_fig:
            fig, ax = ret_val
            return fig

    def predict_video(
        self,
        input_video_path,
        metadata_file,
        threshold=0.5,
        nms_overlap=0.1,
        track=False,
        visualize=False,
        output_file_path=None,
        multiplex=False,
        multiplex_file_path=None,
        tracker_options={
            "assignment_iou_thrd": 0.3,
            "vanish_frames": 40,
            "detect_frames": 10,
        },
        visual_options={
            "show_scores": True,
            "show_labels": True,
            "thickness": 2,
            "fontface": 0,
            "color": (255, 255, 255),
        },
        resize=False,
    ):
        """
        Runs prediction on a video and appends the output VMTI predictions in the metadata file.
        This method is only supported for RGB images.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        input_video_path        Required. Path to the video file to make the
                                predictions on.
        ---------------------   -------------------------------------------
        metadata_file           Required. Path to the metadata csv file where
                                the predictions will be saved in VMTI format.
        ---------------------   -------------------------------------------
        threshold               Optional float. The probability above which
                                a detection will be considered.
        ---------------------   -------------------------------------------
        nms_overlap             Optional float. The intersection over union
                                threshold with other predicted bounding
                                boxes, above which the box with the highest
                                score will be considered a true positive.
        ---------------------   -------------------------------------------
        track                   Optional bool. Set this parameter as True to
                                enable object tracking.
        ---------------------   -------------------------------------------
        visualize               Optional boolean. If True a video is saved
                                with prediction results.
        ---------------------   -------------------------------------------
        output_file_path        Optional path. Path of the final video to be saved.
                                If not supplied, video will be saved at path input_video_path
                                appended with _prediction.
        ---------------------   -------------------------------------------
        multiplex               Optional boolean. Runs Multiplex using the VMTI detections.
        ---------------------   -------------------------------------------
        multiplex_file_path     Optional path. Path of the multiplexed video to be saved.
                                By default a new file with _multiplex.MOV extension is saved
                                in the same folder.
        ---------------------   -------------------------------------------
        tracking_options        Optional dictionary. Set different parameters for
                                object tracking. assignment_iou_thrd parameter is used
                                to assign threshold for assignment of trackers,
                                vanish_frames is the number of frames the object should
                                be absent to consider it as vanished, detect_frames
                                is the number of frames an object should be detected
                                to track it.
        ---------------------   -------------------------------------------
        visual_options          Optional dictionary. Set different parameters for
                                visualization.
                                show_scores boolean, to view scores on predictions,
                                show_labels boolean, to view labels on predictions,
                                thickness integer, to set the thickness level of box,
                                fontface integer, fontface value from opencv values,
                                color tuple (B, G, R), tuple containing values between
                                0-255.
        ---------------------   -------------------------------------------
        resize                  Optional boolean. Resizes the image to the
                                same size (chip_size parameter in prepare_data)
                                that the model was trained on, before detecting
                                objects. Note that if resize_to parameter was
                                used in prepare_data, the image is resized to
                                that size instead.

                                By default, this parameter is false and the
                                detections are run in a sliding window fashion
                                by applying the model on cropped sections of
                                the image (of the same size as the model was
                                trained on).
        =====================   ===========================================

        """

        VideoUtils.predict_video(
            self,
            input_video_path,
            metadata_file,
            threshold,
            nms_overlap,
            track,
            visualize,
            output_file_path,
            multiplex,
            multiplex_file_path,
            tracker_options,
            visual_options,
            resize,
        )

    def _predict_batch(self, images):
        model = self.learn.model
        model.eval()
        model = model.to(self._device)
        normed_batch_tensor = images.to(self._device)
        predictions = model(normed_batch_tensor)
        normed_batch_tensor.detach().cpu()
        del normed_batch_tensor
        return predictions

    def _get_batched_predictions(self, chips, tytx, norm, batch_size=1):
        data = []
        data_counter = 0
        final_class = []
        final_bbox = []
        for idx in range(len(chips)):
            chip = chips[idx]
            frame = np.moveaxis(
                norm(cv2.cvtColor(chip["chip"], cv2.COLOR_BGR2RGB).astype(np.float32)),
                -1,
                0,
            )
            data.append(frame)
            data_counter += 1
            if data_counter % batch_size == 0 or idx == len(chips) - 1:
                batch = chips_to_batch(data, tytx, tytx, batch_size)
                batch_classes, batch_bboxes = self._predict_batch(
                    torch.tensor(batch).float()
                )
                extra_chips = batch_size - len(data)
                batch_output_class = (
                    batch_classes[: (len(batch_classes) - extra_chips)].detach().cpu()
                )
                batch_output_bbox = (
                    batch_bboxes[: (len(batch_bboxes) - extra_chips)].detach().cpu()
                )
                final_class.append(batch_output_class)
                final_bbox.append(batch_output_bbox)
                data = []
                data_counter = 0
        return torch.cat(final_class), torch.cat(final_bbox)

    def predict(
        self,
        image_path,
        threshold=0.5,
        nms_overlap=0.1,
        return_scores=False,
        visualize=False,
        resize=False,
        batch_size=1,
    ):
        """
        Runs prediction on an Image.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        image_path              Required. Path to the image file to make the
                                predictions on.
        ---------------------   -------------------------------------------
        threshold               Optional float. The probability above which
                                a detection will be considered valid.
        ---------------------   -------------------------------------------
        nms_overlap             Optional float. The intersection over union
                                threshold with other predicted bounding
                                boxes, above which the box with the highest
                                score will be considered a true positive.
        ---------------------   -------------------------------------------
        return_scores           Optional boolean. Will return the probability
                                scores of the bounding box predictions if True.
        ---------------------   -------------------------------------------
        visualize               Optional boolean. Displays the image with
                                predicted bounding boxes if True.
        ---------------------   -------------------------------------------
        resize                  Optional boolean. Resizes the image to the
                                same size (chip_size parameter in prepare_data)
                                that the model was trained on, before detecting
                                objects. Note that if resize_to parameter was
                                used in prepare_data, the image is resized to
                                that size instead.

                                By default, this parameter is false and the
                                detections are run in a sliding window fashion
                                by applying the model on cropped sections of
                                the image (of the same size as the model was
                                trained on).
        ---------------------   -------------------------------------------
        batch_size              Optional int. Batch size to be used
                                during tiled inferencing. Deafult value 1.
        =====================   ===========================================

        :return: 'List' of xmin, ymin, width, height of predicted bounding boxes on the given image
        """
        if not HAS_OPENCV:
            raise Exception(
                "This function requires opencv 4.0.1.24. Install it using pip install opencv-python==4.0.1.24"
            )

        if isinstance(image_path, str):
            if self._data._is_multispectral:
                resize_to = None
                if resize:
                    if self._data.resize_to is not None:
                        resize_to = self._data.resize_to
                    elif self._data.chip_size is not None:
                        resize_to = self._data.chip_size
                image = read_image(image_path, resize_to)
            else:
                image = cv2.imread(image_path)
        else:
            image = image_path

        if image is None:
            raise Exception(str("No such file or directory: %s" % (image_path)))
        orig_height, orig_width, _ = image.shape
        orig_frame = image.copy()
        tytx = self._data.chip_size

        if not self._data._is_multispectral:
            if (
                resize
                and self._data.resize_to is None
                and self._data.chip_size is not None
            ):
                image = cv2.resize(image, (self._data.chip_size, self._data.chip_size))

            if self._data.resize_to is not None:
                if isinstance(self._data.resize_to, tuple):
                    image = cv2.resize(image, self._data.resize_to)
                else:
                    image = cv2.resize(
                        image, (self._data.resize_to, self._data.resize_to)
                    )

        height, width, _ = image.shape

        if self._data.chip_size is not None:
            chips = _get_image_chips(image, self._data.chip_size)
        else:
            chips = [
                {
                    "width": width,
                    "height": height,
                    "xmin": 0,
                    "ymin": 0,
                    "chip": image,
                    "predictions": [],
                }
            ]

        include_pad_detections = False
        if len(chips) == 1:
            include_pad_detections = True

        imagenet_stats = ([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])

        mean = 255 * np.array(imagenet_stats[0], dtype=np.float32)
        std = 255 * np.array(imagenet_stats[1], dtype=np.float32)
        norm = lambda x: (x - mean) / std

        valid_tfms = self._data.valid_ds.tfms
        self._data.valid_ds.tfms = []

        from .._utils.pascal_voc_rectangles import modified_getitem
        from fastai.data_block import LabelList

        orig_getitem = LabelList.__getitem__
        LabelList.__getitem__ = modified_getitem

        try:
            pred_class, pred_bbox = self._get_batched_predictions(
                chips, tytx, norm, batch_size
            )

            class dummy:
                pass

            dummy_x = dummy()
            dummy_x.size = [tytx, tytx]

            for chip_idx, (pc, pb) in enumerate(zip(pred_class, pred_bbox)):
                pc = pc.detach().clone()
                pb = pb.detach().clone()
                pp_output = self._analyze_pred(
                    pred=(pc, pb), thresh=threshold, nms_overlap=nms_overlap
                )
                bbox = _reconstruct(
                    pp_output, dummy_x, pad_idx=0, classes=self._data.classes
                )
                if bbox is not None:
                    scores = bbox.scores
                    bboxes, lbls = bbox._compute_boxes()
                    bboxes.add_(1).mul_(
                        torch.tensor(
                            [
                                chips[chip_idx]["height"] / 2,
                                chips[chip_idx]["width"] / 2,
                                chips[chip_idx]["height"] / 2,
                                chips[chip_idx]["width"] / 2,
                            ]
                        )
                    ).long()
                    for index, bbox in enumerate(bboxes):
                        if lbls is not None:
                            label = lbls[index]
                        else:
                            label = "Default"

                        data = bb2hw(bbox)
                        if include_pad_detections or not _exclude_detection(
                            (data[0], data[1], data[2], data[3]),
                            chips[chip_idx]["width"],
                            chips[chip_idx]["height"],
                        ):
                            chips[chip_idx]["predictions"].append(
                                {
                                    "xmin": data[0],
                                    "ymin": data[1],
                                    "width": data[2],
                                    "height": data[3],
                                    "score": float(scores[index]),
                                    "label": label,
                                }
                            )
        finally:
            LabelList.__getitem__ = orig_getitem

        self._data.valid_ds.tfms = valid_tfms

        predictions, labels, scores = _get_transformed_predictions(chips)

        y_ratio = orig_height / height
        x_ratio = orig_width / width

        for index, prediction in enumerate(predictions):
            prediction[0] = prediction[0] * x_ratio
            prediction[1] = prediction[1] * y_ratio
            prediction[2] = prediction[2] * x_ratio
            prediction[3] = prediction[3] * y_ratio

            # Clip xmin
            if prediction[0] < 0:
                prediction[2] = prediction[2] + prediction[0]
                prediction[0] = 1

            # Clip width when xmax greater than original width
            if prediction[0] + prediction[2] > orig_width:
                prediction[2] = (prediction[0] + prediction[2]) - orig_width

            # Clip ymin
            if prediction[1] < 0:
                prediction[3] = prediction[3] + prediction[1]
                prediction[1] = 1

            # Clip height when ymax greater than original height
            if prediction[1] + prediction[3] > orig_height:
                prediction[3] = (prediction[1] + prediction[3]) - orig_height

            predictions[index] = [
                prediction[0],
                prediction[1],
                prediction[2],
                prediction[3],
            ]

        if visualize:
            if self._data._is_multispectral:
                t = torch.tensor(
                    np.rollaxis(orig_frame, -1, 0).astype(np.float32),
                    dtype=torch.float32,
                )[None]
                # im = PIL.Image.fromarray(orig_frame)
                # t = pil2tensor(im, dtype=np.float32)[None]
                scaled_t = self._data._min_max_scaler(t)[0]
                orig_frame = (
                    (scaled_t * 255)
                    .round()
                    .numpy()
                    .astype(np.uint8)[self._data._symbology_rgb_bands]
                )
                orig_frame = np.rollaxis(orig_frame, 0, 3)
                a = np.zeros(orig_frame.shape, dtype=np.uint8)
                a[:] = orig_frame[:]
                if len(labels) > 0:
                    image = _draw_predictions(a, predictions, labels)
                else:
                    image = orig_frame
            else:
                image = _draw_predictions(orig_frame, predictions, labels)
                image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
            import matplotlib.pyplot as plt

            plt.xticks([])
            plt.yticks([])
            plt.imshow(PIL.Image.fromarray(image))

        if return_scores:
            return predictions, labels, scores
        else:
            return predictions, labels

    def average_precision_score(
        self, detect_thresh=0.2, iou_thresh=0.1, mean=False, show_progress=True
    ):
        """
        Computes average precision on the validation set for each class.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        detect_thresh           Optional float. The probability above which
                                a detection will be considered for computing
                                average precision.
        ---------------------   -------------------------------------------
        iou_thresh              Optional float. The intersection over union
                                threshold with the ground truth labels, above
                                which a predicted bounding box will be
                                considered a true positive.
        ---------------------   -------------------------------------------
        mean                    Optional bool. If False returns class-wise
                                average precision otherwise returns mean
                                average precision.
        =====================   ===========================================

        :return: `dict` if mean is False otherwise `float`
        """
        self._check_requisites()

        aps = compute_class_AP(
            self,
            self._data.valid_dl,
            self._data.c - 1,
            show_progress,
            detect_thresh=detect_thresh,
            iou_thresh=iou_thresh,
        )
        if mean:
            import statistics

            return statistics.mean(aps)
        else:
            return dict(zip(self._data.classes[1:], aps))

    ## Tensorflow specific functions start ##
    def _intialize_tensorflow(
        self,
        data,
        grids,
        zooms,
        ratios,
        backbone,
        drop,
        bias,
        pretrained_path,
        location_loss_factor,
    ):
        self._check_tf()

        from .._utils.fastai_tf_fit import TfLearner
        import tensorflow as tf
        from tensorflow.keras.losses import BinaryCrossentropy
        from tensorflow.keras.models import Model
        from tensorflow.keras import applications
        from tensorflow.keras.optimizers import Adam
        from fastai.basics import defaults
        from .._utils.object_detection import get_ssd_head_output

        if data._is_multispectral:
            raise Exception(
                'Multispectral data is not supported with backend="tensorflow"'
            )

        # prepare color array
        alpha = 0.7
        color_mapping = getattr(data, "color_mapping", None)
        if color_mapping is None:
            color_array = torch.tensor([[1.0, 1.0, 1.0]]).float()
        else:
            color_array = torch.tensor(list(color_mapping.values())).float() / 255
        alpha_tensor = torch.tensor([alpha] * len(color_array)).view(-1, 1).float()
        color_array = torch.cat([color_array, alpha_tensor], dim=-1)
        background_color = torch.tensor([[0, 0, 0, 0]]).float()
        data._multispectral_color_array = torch.cat([background_color, color_array])

        self.ssd_version = 1  # ssd_version
        if backbone is None:
            backbone = "ResNet50"

        if type(backbone) == str:
            backbone = getattr(applications, backbone)

        self._backbone = backbone

        x, y = next(iter(data.train_dl))
        if tf.keras.backend.image_data_format() == "channels_last":
            in_shape = [x.shape[-1], x.shape[-1], 3]
        else:
            in_shape = [3, x.shape[-1], x.shape[-1]]

        self._backbone_initalized = self._backbone(
            input_shape=in_shape,
            include_top=False,
            # weights='imagenet'
        )
        self._backbone_initalized.trainable = False

        self._device = torch.device("cpu")
        self._data = data

        # self._loss_function_classification = BinaryCrossentropy(from_logits=True, reduction=Reduction.SUM) #2.0.0
        self._loss_function_classification = BinaryCrossentropy(
            from_logits=True, reduction="sum"
        )
        self.location_loss_factor = location_loss_factor

        if grids is None:
            # find most suitable centroids for dataset
            height_width = np.array(data.height_width)
            centroid = kmeans(height_width, 1)
            avg = avg_iou(height_width, centroid)

            for num_anchor in range(2, 5):
                new_centroid = kmeans(height_width, num_anchor)
                new_avg = avg_iou(height_width, new_centroid)
                if (new_avg - avg) < 0.05:
                    break
                avg = new_avg
                centroid = new_centroid.copy()

            # find grid size
            grids = list(
                map(int, map(round, data.chip_size / np.sort(np.max(centroid, axis=1))))
            )

            grids = list(set(grids))
            grids.sort(reverse=True)
            if grids[-1] == 0:
                grids[-1] = 1
            grids = list(set(grids))

        self.grids = grids
        self.zooms = zooms
        self.ratios = ratios

        self._create_anchors(grids, zooms, ratios)

        output_layer = get_ssd_head_output(self)

        model = Model(inputs=self._backbone_initalized.input, outputs=output_layer)

        self.learn = TfLearner(
            data,
            model,
            opt_func=Adam,
            loss_func=self._loss_func_tf,
            true_wd=True,
            bn_wd=True,
            wd=defaults.wd,
            train_bn=True,
        )

        self.learn.unfreeze()
        self.learn.freeze_to(len(self._backbone_initalized.layers))

        self.show_results = self._show_results_multispectral

        self._code = code
        if pretrained_path is not None:
            self.load(pretrained_path)

    def _loss_func_tf(self, y_bboxes, y_classes, predictions):
        from .._utils.object_detection import tf_loss_function_single_image
        import tensorflow as tf

        predicted_classes = predictions[0]
        predicted_bboxes = predictions[1]

        localization_loss, classification_loss = tf.constant(0.0), tf.constant(0.0)
        # Now we will iterate over each image and calculate loss for a single image
        for image_y_bboxes, image_y_calsses, image_p_bboxes, image_p_classes in zip(
            y_bboxes, y_classes, predicted_bboxes, predicted_classes
        ):
            _localization_loss, _classification_loss = tf_loss_function_single_image(
                self, image_y_bboxes, image_y_calsses, image_p_bboxes, image_p_classes
            )
            classification_loss += _classification_loss
            localization_loss += _localization_loss

        if self.location_loss_factor is None:
            return localization_loss + classification_loss
        else:
            return (
                self.location_loss_factor * localization_loss
                + (1 - self.location_loss_factor) * classification_loss
            )

    ## Tensorflow specific functions end ##



from distutils.version import LooseVersion
import json, logging
import warnings

from pandas import DataFrame
import numpy as np
from arcgis import __version__
from arcgis.geometry import Geometry, SpatialReference

import pyarrow
import pyarrow.feather as feather
import pyarrow.parquet as parquet
from arcgis.geometry._types import _check_geometry_engine
from arcgis.features.geo._array import GeoArray

_logging = logging.getLogger()

_HASARCPY, _HASSHAPELY = _check_geometry_engine()

_METADATA_VERSION = "0.4.0"
# reference: https://github.com/geopandas/geo-arrow-spec

# Metadata structure:
# {
#     "geo": {
#         "columns": {
#             "<name>": {
#                 "crs": "<WKT or None: REQUIRED>",
#                 "encoding": "WKB",
#                 "geometry_tye" : "<OGC Types>",
#                 "bbox" : [xmin, ymin, xmax, ymax]
#             }
#         },
#         "creator": {
#             "library": "arcgis",
#             "version": "<arcgis.__version__>"
#         }
#         "primary_column": "<str: REQUIRED>",
#         "schema_version": "<METADATA_VERSION>"
#     }
# }


def _create_metadata(df):
    """Create and encode geo metadata dict.

    Parameters
    ----------
    df : pd.DataFrame

    Returns
    -------
    dict
    """
    _gt_lu = {
        "point": "Point",
        "polyline": "MultiLineString",
        "polygon": "MultiPolygon",
        "multipoint": "MultiPoint",
        "unknown": "Unknown",
        "extent": "Unknown",
    }
    # Construct metadata for each geometry
    column_metadata = {}
    for col in df.columns[df.dtypes == "geometry"]:
        # series = df[col]

        if _HASARCPY:
            sr = SpatialReference(df.spatial.sr).as_arcpy.exportToString()
        else:
            sr = f"ESPG:{df.spatial.sr.get('wkid', 4326)}"
        gt = [_gt_lu[g.lower()] for g in df.spatial.geometry_type]
        if len(gt) == 1:
            gt = gt[0]
        column_metadata[col] = {
            "crs": sr,  # df.spatial.sr[0],#series.crs.to_wkt() if series.crs else None,
            "encoding": "WKB",
            "bbox": df.spatial.full_extent,
            "geometry_type": gt,
        }

    return {
        "primary_column": df.spatial.name,
        "columns": column_metadata,
        "schema_version": _METADATA_VERSION,
        "creator": {"library": "arcgis", "version": __version__},
    }


def _encode_metadata(metadata):
    """Encode metadata dict to UTF-8 JSON string

    Parameters
    ----------
    metadata : dict

    Returns
    -------
    UTF-8 encoded JSON string
    """
    return json.dumps(metadata).encode("utf-8")


def _decode_metadata(metadata_str):
    """Decode a UTF-8 encoded JSON string to dict

    Parameters
    ----------
    metadata_str : string (UTF-8 encoded)

    Returns
    -------
    dict
    """
    if metadata_str is None:
        return None

    return json.loads(metadata_str.decode("utf-8"))


def _validate_dataframe(df):
    """Validate that the Spatially Enabled DataFrame conforms to requirements for writing
    to Parquet format.

    Raises `ValueError` if the Spatially Enabled DataFrame is not valid.

    copied from `pandas.io.parquet`

    Parameters
    ----------
    df : pd.DataFrame
    """

    if not isinstance(df, DataFrame):
        raise ValueError("Writing to Parquet/Feather only supports IO with DataFrames")

    # must have value column names (strings only)
    if df.columns.inferred_type not in {"string", "unicode", "empty"}:
        raise ValueError("Writing to Parquet/Feather requires string column names")

    # index level names must be strings
    valid_names = all(
        isinstance(name, str) for name in df.index.names if name is not None
    )
    if not valid_names:
        raise ValueError("Index level names must be strings")


def _validate_metadata(metadata):
    """Validate geo metadata.
    Must not be empty, and must contain the structure specified above.

    Raises ValueError if metadata is not valid.

    Parameters
    ----------
    metadata : dict
    """

    if not metadata:
        raise ValueError("Missing or malformed geo metadata in Parquet/Feather file")

    required_keys = ("primary_column", "columns")
    for key in required_keys:
        if metadata.get(key, None) is None:
            raise ValueError(
                "'geo' metadata in Parquet/Feather file is missing required key: "
                "'{key}'".format(key=key)
            )

    if not isinstance(metadata["columns"], dict):
        raise ValueError("'columns' in 'geo' metadata must be a dict")

    # Validate that geometry columns have required metadata and values
    required_col_keys = ("crs", "encoding")
    for col, column_metadata in metadata["columns"].items():
        for key in required_col_keys:
            if key not in column_metadata:
                raise ValueError(
                    "'geo' metadata in Parquet/Feather file is missing required key "
                    "'{key}' for column '{col}'".format(key=key, col=col)
                )

        if column_metadata["encoding"] != "WKB":
            raise ValueError("Only WKB geometry encoding is supported")


def _sedf_to_arrow(df, index=None):
    """
    Helper function with main, shared logic for to_parquet/to_feather.
    """
    from pyarrow import Table

    _validate_dataframe(df)

    # create geo metadata before altering incoming data frame
    geo_metadata = _create_metadata(df)

    table = Table.from_pandas(df, preserve_index=index)

    # Store geopandas specific file-level metadata
    # This must be done AFTER creating the table or it is not persisted
    metadata = table.schema.metadata
    metadata.update({b"geo": _encode_metadata(geo_metadata)})
    fin = table.replace_schema_metadata(metadata)

    return fin


def _to_parquet(
    df: "DataFrame", path: str, index: bool = None, compression: str = "gzip", **kwargs
) -> str:
    """
    Write a Spatially Enabled DataFrame to the Parquet format.

    Any geometry columns present are serialized to WKB format in the file.

    Requires 'pyarrow'.

    WARNING: this is an initial implementation of Parquet file support and
    associated metadata.  This is tracking version 0.1.0 of the metadata
    specification at:
    https://github.com/geopandas/geo-arrow-spec

    .. versionadded:: 1.9

    Parameters
    ----------
    path : str, path object
    index : bool, default None
        If ``True``, always include the dataframe's index(es) as columns
        in the file output.
        If ``False``, the index(es) will not be written to the file.
        If ``None``, the index(ex) will be included as columns in the file
        output except `RangeIndex` which is stored as metadata only.
    compression : {'snappy', 'gzip', 'brotli', None}, default 'snappy'
        Name of the compression to use. Use ``None`` for no compression.
    kwargs
        Additional keyword arguments passed to pyarrow.parquet.write_table().
    """
    table = _sedf_to_arrow(df, index=index)
    parquet.write_table(table, path, compression=compression, **kwargs)
    return path


def _to_feather(df, path, index=None, compression=None, **kwargs) -> str:
    """
    Write a Spatially Enabled DataFrame to the Feather format.

    Any geometry columns present are serialized to WKB format in the file.

    Requires 'pyarrow' >= 0.17.

    WARNING: this is an initial implementation of Feather file support and
    associated metadata.  This is tracking version 0.1.0 of the metadata
    specification at:
    https://github.com/geopandas/geo-arrow-spec

    .. versionadded:: 1.9

    Parameters
    ----------
    path : str, path object
    index : bool, default None
        If ``True``, always include the dataframe's index(es) as columns
        in the file output.
        If ``False``, the index(es) will not be written to the file.
        If ``None``, the index(ex) will be included as columns in the file
        output except `RangeIndex` which is stored as metadata only.
    compression : {'zstd', 'lz4', 'uncompressed'}, optional
        Name of the compression to use. Use ``"uncompressed"`` for no
        compression. By default uses LZ4 if available, otherwise uncompressed.
    kwargs
        Additional keyword arguments passed to pyarrow.feather.write_feather().
    """

    if pyarrow.__version__ < LooseVersion("0.17.0"):
        raise ImportError("pyarrow >= 0.17 required for Feather support")

    table = _sedf_to_arrow(df, index=index)
    feather.write_feather(table, path, compression=compression, **kwargs)
    return path


def _arrow_to_sedf(table) -> "pandas.DataFrame":
    """
    Helper function with main, shared logic for read_parquet/read_feather.
    """
    df = table.to_pandas()

    metadata = table.schema.metadata
    if b"geo" not in metadata:
        raise ValueError(
            """Missing geo metadata in Parquet/Feather file.
            Use pandas.read_parquet/read_feather() instead."""
        )

    try:
        metadata = _decode_metadata(metadata.get(b"geo", b""))

    except (TypeError, json.decoder.JSONDecodeError):
        raise ValueError("Missing or malformed geo metadata in Parquet/Feather file")

    _validate_metadata(metadata)

    # Find all geometry columns that were read from the file.  May
    # be a subset if 'columns' parameter is used.
    geometry_columns = df.columns.intersection(metadata["columns"])

    if not len(geometry_columns):
        raise ValueError(
            """No geometry columns are included in the columns read from
            the Parquet/Feather file.  To read this file without geometry columns,
            use pandas.read_parquet/read_feather() instead."""
        )

    geometry = metadata["primary_column"]

    # Missing geometry likely indicates a subset of columns was read;
    # promote the first available geometry to the primary geometry.
    if len(geometry_columns) and geometry not in geometry_columns:
        geometry = geometry_columns[0]

        # if there are multiple non-primary geometry columns, raise a warning
        if len(geometry_columns) > 1:
            _logging.warning(
                "Multiple non-primary geometry columns read from Parquet/Feather "
                "file. The first column read was promoted to the primary geometry."
            )

    # Convert the WKB columns that are present back to geometry.

    if _HASARCPY:
        for col in geometry_columns:
            array = np.empty(len(table.column(geometry)), "O")
            array[:] = [Geometry(i.as_py()) for i in table.column(geometry)]
            df[col] = array
            df.spatial.set_geometry(geometry)
            for key in metadata["columns"].keys():
                data = metadata["columns"][key]
                if key == metadata["primary_column"]:
                    if (
                        "crs" in data
                        and data["crs"]
                        and data["crs"].find("ESPG:") == -1
                    ):
                        df.spatial.sr = {"wkt": data["crs"]}
                    elif (
                        "crs" in data and data["crs"] and data["crs"].find("ESPG:") > -1
                    ):
                        df.spatial.sr = {"wkid": int(data["crs"].split(":")[0])}
                    else:
                        df.spatial.sr = {"wkid": 4326}
                else:
                    df.spatial.name = key
                    if (
                        "crs" in data
                        and data["crs"]
                        and data["crs"].find("ESPG:") == -1
                    ):
                        df.spatial.sr = {"wkt": data["crs"]}
                    elif (
                        "crs" in data and data["crs"] and data["crs"].find("ESPG:") > -1
                    ):
                        df.spatial.sr = {"wkid": int(data["crs"].split(":")[0])}
                    else:
                        df.spatial.sr = {"wkid": 4326}
            if df.spatial.name != geometry:
                df.spatial.set_geometry(geometry)
    else:
        from geomet import wkb
        from geomet import esri as _esri

        for col in geometry_columns:
            array = np.empty(len(table.column(geometry)), "O")
            array[:] = [
                Geometry(_esri.dumps(wkb.loads(i.as_py())))
                for i in table.column(geometry)
            ]
            df[col] = array
            df.spatial.set_geometry(geometry)
    return df


def _read_parquet(path, columns=None, **kwargs):
    """
    Load a Parquet object from the file path, returning a Spatially Enabled DataFrame.

    You can read a subset of columns in the file using the ``columns`` parameter.
    However, the structure of the returned Spatially Enabled DataFrame will depend on which
    columns you read:

    * if no geometry columns are read, this will raise a ``ValueError`` - you
      should use the pandas `read_parquet` method instead.
    * if the primary geometry column saved to this file is not included in
      columns, the first available geometry column will be set as the geometry
      column of the returned Spatially Enabled DataFrame.

    Requires 'pyarrow'.

    .. versionadded:: 1.9

    Parameters
    ----------
    path : str, path object
    columns : list-like of strings, default=None
        If not None, only these columns will be read from the file.  If
        the primary geometry column is not included, the first secondary
        geometry read from the file will be set as the geometry column
        of the returned Spatially Enabled DataFrame.  If no geometry columns are present,
        a ``ValueError`` will be raised.
    **kwargs
        Any additional kwargs passed to pyarrow.parquet.read_table().

    Returns
    -------
    Spatially Enabled DataFrame

    Examples
    --------
    >>> df = pd.DataFrame.spatial.read_parquet("data.parquet")  # doctest: +SKIP

    Specifying columns to read:

    >>> df = pd.DataFrame.spatial.read_parquet(
    ...     "data.parquet",
    ...     columns=["geometry", "pop_est"]
    ... )  # doctest: +SKIP
    """
    kwargs["use_pandas_metadata"] = True
    table = parquet.read_table(path, columns=columns, **kwargs)

    return _arrow_to_sedf(table)


def _read_feather(path, columns=None, **kwargs):
    """
    Load a Feather object from the file path, returning a Spatially Enabled DataFrame.

    You can read a subset of columns in the file using the ``columns`` parameter.
    However, the structure of the returned Spatially Enabled DataFrame will depend on which
    columns you read:

    * if no geometry columns are read, this will raise a ``ValueError`` - you
      should use the pandas `read_feather` method instead.
    * if the primary geometry column saved to this file is not included in
      columns, the first available geometry column will be set as the geometry
      column of the returned Spatially Enabled DataFrame.

    Requires 'pyarrow' >= 0.17.

    .. versionadded:: 1.9

    Parameters
    ----------
    path : str, path object
    columns : list-like of strings, default=None
        If not None, only these columns will be read from the file.  If
        the primary geometry column is not included, the first secondary
        geometry read from the file will be set as the geometry column
        of the returned Spatially Enabled DataFrame.  If no geometry columns are present,
        a ``ValueError`` will be raised.
    **kwargs
        Any additional kwargs passed to pyarrow.feather.read_table().

    Returns
    -------
    Spatially Enabled DataFrame

    Examples
    --------
    >>> df = pd.DataFrame.spatial.read_feather("data.feather")  # doctest: +SKIP

    Specifying columns to read:

    >>> df = pd.DataFrame.spatial.read_feather(
    ...     "data.feather",
    ...     columns=["geometry", "pop_est"]
    ... )  # doctest: +SKIP
    """

    if pyarrow.__version__ < LooseVersion("0.17.0"):
        raise ImportError("pyarrow >= 0.17 required for Feather support")

    table = feather.read_table(path, columns=columns, **kwargs)
    return _arrow_to_sedf(table)



# Copyright 2018-2019 Open-MMLab.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pisa_faster_rcnn_x101_32x4d_fpn_1x_coco.py, box AP=41.9
_base_ = "./_base_/models/faster_rcnn_r50_fpn.py"

model = dict(
    pretrained="open-mmlab://resnext101_32x4d",
    backbone=dict(
        type="ResNeXt",
        depth=101,
        groups=32,
        base_width=4,
        num_stages=4,
        out_indices=(0, 1, 2, 3),
        frozen_stages=1,
        norm_cfg=dict(type="BN", requires_grad=True),
        style="pytorch",
    ),
    roi_head=dict(
        type="PISARoIHead",
        bbox_head=dict(loss_bbox=dict(type="SmoothL1Loss", beta=1.0, loss_weight=1.0)),
    ),
    train_cfg=dict(
        rpn_proposal=dict(
            nms_across_levels=False,
            nms_pre=2000,
            nms_post=2000,
            max_per_img=2000,
            nms=dict(type="nms", iou_threshold=0.7),
            min_bbox_size=0,
        ),
        rcnn=dict(
            sampler=dict(
                type="ScoreHLRSampler",
                num=512,
                pos_fraction=0.25,
                neg_pos_ub=-1,
                add_gt_as_proposals=True,
                k=0.5,
                bias=0.0,
            ),
            isr=dict(k=2, bias=0),
            carl=dict(k=1, bias=0.2),
        ),
    ),
    test_cfg=dict(
        rpn=dict(
            nms_across_levels=False,
            nms_pre=2000,
            nms_post=2000,
            max_per_img=2000,
            nms=dict(type="nms", iou_threshold=0.7),
            min_bbox_size=0,
        )
    ),
)

checkpoint = "http://download.openmmlab.com/mmdetection/v2.0/pisa/pisa_faster_rcnn_x101_32x4d_fpn_1x_coco/pisa_faster_rcnn_x101_32x4d_fpn_1x_coco-e4accec4.pth"



from enum import unique
from pathlib import Path
import json
from .._data import prepare_data

import numpy
from ._model_extension import ModelExtension
from ._arcgis_model import _EmptyData
import logging

from .._mmseg_config.prithvi100m_burn_scar import img_norm_burn_model
from .._mmseg_config.prithvi100m_crop_classification import img_norm_crop_model
from .._mmseg_config.prithvi100m_sen1floods import img_norm_flood_model

logger = logging.getLogger()

try:
    from fastai.vision import flatten_model
    import torch
    from fastai.torch_core import split_model_idx
    from .._utils.common import get_multispectral_data_params_from_emd, _get_emd_path

    HAS_FASTAI = True

except Exception as e:
    HAS_FASTAI = False


class MMSegmentationConfig:
    """
    Create class with following fixed function names and the number of arguents to train your model from external source
    """

    try:
        import torch
        import numpy
        import pathlib
        import os
        import types
    except:
        pass

    def get_model(self, data, backbone=None, **kwargs):
        """
        In this fuction you have to define your model with following two arguments!

        """

        import mmseg.models
        import mmcv
        import logging

        logging.disable(logging.WARNING)

        config = kwargs.get("model", False)
        checkpoint = kwargs.get("model_weight", False)
        class_weight = kwargs.get("class_weight", None)
        if config[-2:] != "py":
            config += ".py"
        if self.os.path.exists(self.pathlib.Path(config)):
            cfg = mmcv.Config.fromfile(config)
            cfg.model.pretrained = None

            # changes normalizaion layers for custom cfg since by default mmseg config consider multigpu env
            def change_norm_layer(cfg):
                for k, v in cfg.items():
                    if k == "norm_cfg":
                        cfg[k].type = "BN"
                    elif isinstance(cfg[k], dict):
                        change_norm_layer(cfg[k])

            change_norm_layer(cfg.model)
        else:
            import arcgis

            cfg_abs_path = (
                self.pathlib.Path(arcgis.__file__).parent
                / "learn"
                / "_mmseg_config"
                / config
            )
            cfg = mmcv.Config.fromfile(cfg_abs_path)
            checkpoint = cfg.get("checkpoint", False)
            if checkpoint:
                cfg.model.pretrained = None

        if isinstance(cfg.model.decode_head, list):
            for dcd_head in cfg.model.decode_head:
                dcd_head.num_classes = data.c
                dcd_head.loss_decode.class_weight = class_weight
        else:
            cfg.model.decode_head.num_classes = data.c
            cfg.model.decode_head.loss_decode.class_weight = class_weight

        if hasattr(cfg.model, "auxiliary_head"):
            if isinstance(cfg.model.auxiliary_head, list):
                for aux_head in cfg.model.auxiliary_head:
                    aux_head.num_classes = data.c
                    aux_head.loss_decode.class_weight = class_weight
            else:
                cfg.model.auxiliary_head.num_classes = data.c
                cfg.model.auxiliary_head.loss_decode.class_weight = class_weight
        if cfg.model.backbone.type == "CGNet" and getattr(
            data, "_is_multispectral", False
        ):
            cfg.model.backbone.in_channels = len(data._extract_bands)

        import warnings

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            model = mmseg.models.build_segmentor(cfg.model)

        if checkpoint:
            mmcv.runner.load_checkpoint(
                model, checkpoint, "cpu", False, logging.getLogger()
            )

        # default forward of the model from the original API should be modified to make it compatible with the learn module.
        from mmcv.runner import auto_fp16

        @auto_fp16(apply_to=("img",))
        def forward_modified(self, img, img_metas=None, gt_semantic_seg=None):
            import warnings

            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                if self.training:
                    losses = self.forward_train(img, img_metas, gt_semantic_seg)
                    loss, log_vars = self._parse_losses(losses)

                    outputs = dict(loss=loss, log_vars=log_vars)

                    return outputs
                else:
                    return self.forward_test(img[0], img[1], rescale=True)

        # default simple_test of the model from the original API should be modified to correctly work in test time.
        def simple_test_modified(self, img, img_meta, rescale=True):
            seg_logit = self.encode_decode(img, img_meta)
            return seg_logit

        model.forward = self.types.MethodType(forward_modified, model)
        model.simple_test = self.types.MethodType(simple_test_modified, model)

        self.model = model
        self.cfg = cfg

        logging.disable(0)

        return model

    def on_batch_begin(self, learn, model_input_batch, model_target_batch, **kwargs):
        image_pad_shape = model_input_batch.permute(0, 2, 3, 1).shape[1:]
        image_scale_factor = self.numpy.array(
            [1.0, 1.0, 1.0, 1.0], dtype=self.numpy.float32
        )

        img_metas_dict = {}
        img_metas_dict["pad_shape"] = image_pad_shape
        img_metas_dict["img_shape"] = image_pad_shape
        img_metas_dict["ori_shape"] = image_pad_shape
        img_metas_dict["scale_factor"] = image_scale_factor
        img_metas_dict["flip"] = False

        img_metas = [img_metas_dict] * model_input_batch.shape[0]

        if learn.model.training:
            model_input = [model_input_batch, img_metas, model_target_batch]
            return model_input, model_target_batch
        else:
            model_input = [[[model_input_batch], [img_metas]]]
            return model_input, model_target_batch

    def transform_input(self, xb):
        image_pad_shape = xb.permute(0, 2, 3, 1).shape[1:]
        image_scale_factor = self.numpy.array(
            [1.0, 1.0, 1.0, 1.0], dtype=self.numpy.float32
        )
        img_metas_dict = {}
        img_metas_dict["pad_shape"] = image_pad_shape
        img_metas_dict["img_shape"] = image_pad_shape
        img_metas_dict["ori_shape"] = image_pad_shape
        img_metas_dict["scale_factor"] = image_scale_factor
        img_metas_dict["flip"] = False

        img_metas = [img_metas_dict] * xb.shape[0]
        model_input = [[xb], [img_metas]]

        return model_input

    def transform_input_multispectral(self, xb):
        return self.transform_input(xb)

    def loss(self, model_output, *model_target):
        if not self.model.training:
            if self.cfg.model.type == "CascadeEncoderDecoder":
                losses = 0.0
                for i in range(self.cfg.model.num_stages):
                    _losses = self.model.decode_head[i].losses(
                        model_output, model_target[0]
                    )
                    losses += _losses.get("loss_ce", _losses.get("loss_seg"))
                return losses

            _losses = self.model.decode_head.losses(model_output, model_target[0])
            loss_dice = _losses.get("loss_dice")
            if loss_dice:
                return loss_dice
            else:
                return _losses.get("loss_ce", _losses.get("loss_seg"))

        return model_output["loss"]

    def post_process(self, pred, thres=0.5, thinning=True, prob_raster=False):
        """
        In this function you have to return list with appended output for each image in the batch with shape [C=1,H,W]!
        """
        if prob_raster:
            return pred
        else:
            pred = self.torch.unsqueeze(pred.argmax(dim=1), dim=1)
        return pred


def norm_prithvi(data, model):
    scaling_info = {
        "prithvi100m_burn_scar": (
            img_norm_burn_model.get("means"),
            img_norm_burn_model.get("stds"),
        ),
        "prithvi100m_sen1floods": (
            img_norm_flood_model.get("means"),
            img_norm_flood_model.get("stds"),
        ),
        "prithvi100m_crop_classification": (
            img_norm_crop_model.get("means"),
            img_norm_crop_model.get("stds"),
        ),
        "prithvi100m": (data._scaled_mean_values, data._scaled_std_values),
    }

    means, stds = scaling_info[model]
    data._scaled_mean_values, data._scaled_std_values = torch.tensor(
        means
    ), torch.tensor(stds)

    data._min_max_scaler = None

    if (data._band_max_values.mean() > 1) and (
        model != "prithvi100m_crop_classification"
    ):
        div_value = 10000
    else:
        div_value = None

    data.valid_ds.x._div = div_value
    data.train_ds.x._div = div_value

    data = data.normalize(
        stats=(data._scaled_mean_values, data._scaled_std_values), do_x=True, do_y=False
    )

    return data


class MMSegmentation(ModelExtension):
    """
    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    data                    Required fastai Databunch. Returned data object from
                            :meth:`~arcgis.learn.prepare_data`  function.
    ---------------------   -------------------------------------------
    model                   Required model name or path to the configuration file
                            from :class:`~arcgis.learn.MMSegmentation` repository. The list of the
                            supported models can be queried using
                            :attr:`~arcgis.learn.MMSegmentation.supported_models`
    ---------------------   -------------------------------------------
    model_weight            Optional path of the model weight from
                            :class:`~arcgis.learn.MMSegmentation` repository.
    ---------------------   -------------------------------------------
    pretrained_path         Optional string. Path where pre-trained model is
                            saved.
    =====================   ===========================================

    **kwargs**

    =====================   ===========================================
    class_balancing         Optional boolean. If True, it will balance the
                            cross-entropy loss inverse to the frequency
                            of pixels per class. Default: False.
    ---------------------   -------------------------------------------
    ignore_classes          Optional list. It will contain the list of class
                            values on which model will not incur loss.
                            Default: []
    =====================   ===========================================

    :return: :class:`~arcgis.learn.MMSegmentation` Object
    """

    def __init__(self, data, model, model_weight=False, pretrained_path=None, **kwargs):
        self._check_dataset_support(data)

        if model.startswith("prithvi100m"):
            data.remove_tfm(data.norm)
            data.norm, data.denorm = None, None
            data = norm_prithvi(data, model)
        self._ignore_classes = kwargs.get("ignore_classes", [])
        self.class_balancing = kwargs.get("class_balancing", False)
        if self._ignore_classes != [] and len(data.classes) <= 2:
            raise Exception(
                f"`ignore_classes` parameter can only be used when the dataset has more than 2 classes."
            )

        data_classes = list(data.class_mapping.keys())
        if 0 not in list(data.class_mapping.values()):
            self._ignore_mapped_class = [
                data_classes.index(k) + 1 for k in self._ignore_classes if k != 0
            ]
        else:
            self._ignore_mapped_class = [
                data_classes.index(k) + 1 for k in self._ignore_classes
            ]
        if self._ignore_classes != []:
            if 0 not in self._ignore_mapped_class:
                self._ignore_mapped_class.insert(0, 0)

        class_weight = None
        if self.class_balancing:
            if data.class_weight is not None:
                # Handle condition when nodata is already at pixel value 0 in data
                if (data.c - 1) == data.class_weight.shape[0]:
                    class_weight = [
                        data.class_weight.mean()
                    ] + data.class_weight.tolist()
                else:
                    class_weight = data.class_weight.tolist()
            else:
                if getattr(data, "overflow_encountered", False):
                    logger.warning(
                        "Overflow Encountered. Ignoring `class_balancing` parameter."
                    )
                    class_weight = [1.0] * len(data.classes)
                else:
                    logger.warning(
                        "Could not find 'NumPixelsPerClass' in 'esri_accumulated_stats.json'. Ignoring `class_balancing` parameter."
                    )

        if self._ignore_classes != []:
            if not self.class_balancing:
                class_weight = [1.0] * data.c
            for idx in self._ignore_mapped_class:
                class_weight[idx] = 0.0

        self._final_class_weight = class_weight

        super().__init__(
            data,
            MMSegmentationConfig,
            pretrained_path=pretrained_path,
            model=model,
            model_weight=model_weight,
            ignore_class=self._ignore_mapped_class,
            class_weight=self._final_class_weight,
        )
        idx = self._freeze()
        self.learn.layer_groups = split_model_idx(self.learn.model, [idx])
        self.learn.create_opt(lr=3e-3)

    def unfreeze(self):
        for _, param in self.learn.model.named_parameters():
            param.requires_grad = True

    def _freeze(self):
        "Freezes the pretrained backbone."
        if self._model_conf.cfg.model.backbone.type == "CGNet":
            return 6
        for idx, i in enumerate(flatten_model(self.learn.model.backbone)):
            if isinstance(i, (torch.nn.BatchNorm2d)):
                continue
            for p in i.parameters():
                p.requires_grad = False
        return idx

    @staticmethod
    def _available_metrics():
        return ["valid_loss", "accuracy"]

    @property
    def _is_mmsegdet(self):
        return True

    @property
    def supported_datasets(self):
        """Supported dataset types for this model."""
        return MMSegmentation._supported_datasets()

    @staticmethod
    def _supported_datasets():
        return ["Classified_Tiles"]

    supported_models = [
        "ann",
        "apcnet",
        "ccnet",
        "cgnet",
        "deeplabv3",
        "deeplabv3plus",
        "dmnet",
        "dnlnet",
        "emanet",
        "fastscnn",
        "fcn",
        "gcnet",
        "hrnet",
        "mobilenet_v2",
        "nonlocal_net",
        "ocrnet",
        "psanet",
        "pspnet",
        "resnest",
        "sem_fpn",
        "unet",
        "upernet",
    ]
    """
    List of models supported by this class.
    """

    @classmethod
    def from_model(cls, emd_path, data=None):
        """
        Creates a :class:`~arcgis.learn.MMSegmentation` object from an Esri Model Definition (EMD) file.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to Deep Learning Package
                                (DLPK) or Esri Model Definition(EMD) file.
        ---------------------   -------------------------------------------
        data                    Required fastai Databunch or None. Returned data
                                object from :meth:`~arcgis.learn.prepare_data`  function or None for
                                inferencing.

        =====================   ===========================================

        :return: :class:`~arcgis.learn.MMSegmentation` Object
        """
        emd_path = _get_emd_path(emd_path)

        with open(emd_path) as f:
            emd = json.load(f)

        model_file = Path(emd["ModelFile"])

        if not model_file.is_absolute():
            model_file = emd_path.parent / model_file

        kwargs = emd.get("Kwargs", {})

        try:
            class_mapping = {i["Value"]: i["Name"] for i in emd["Classes"]}
            color_mapping = {i["Value"]: i["Color"] for i in emd["Classes"]}
        except KeyError:
            class_mapping = {i["ClassValue"]: i["ClassName"] for i in emd["Classes"]}
            color_mapping = {i["ClassValue"]: i["Color"] for i in emd["Classes"]}

        if data is None:
            data = _EmptyData(
                path=emd_path.parent.parent,
                loss_func=None,
                c=len(class_mapping) + 1,
                chip_size=emd["ImageHeight"],
            )
            data.class_mapping = class_mapping
            data.color_mapping = color_mapping
            data._is_empty = True
            data.emd_path = emd_path
            data.emd = emd
            data.classes = ["background"]
            for k, v in class_mapping.items():
                data.classes.append(v)
            data = get_multispectral_data_params_from_emd(data, emd)
            data.dataset_type = emd["DatasetType"]

        return cls(data, pretrained_path=str(model_file), **kwargs)

    def show_results(self, rows=5, thresh=0.5, thinning=True, **kwargs):
        """
        Displays the results of a trained model on a part of the validation set.
        """



import math
import torch
import functools
from torch import nn, einsum
import torch.nn.functional as F
from inspect import isfunction
from functools import partial
import numpy as np
from tqdm import tqdm


def _warmup_beta(linear_start, linear_end, n_timestep, warmup_frac):
    betas = linear_end * np.ones(n_timestep, dtype=np.float64)
    warmup_time = int(n_timestep * warmup_frac)
    betas[:warmup_time] = np.linspace(
        linear_start, linear_end, warmup_time, dtype=np.float64
    )
    return betas


def make_beta_schedule(
    schedule, n_timestep, linear_start=1e-4, linear_end=2e-2, cosine_s=8e-3
):
    if schedule == "quad":
        betas = (
            np.linspace(
                linear_start**0.5, linear_end**0.5, n_timestep, dtype=np.float64
            )
            ** 2
        )
    elif schedule == "linear":
        betas = np.linspace(linear_start, linear_end, n_timestep, dtype=np.float64)
    elif schedule == "warmup10":
        betas = _warmup_beta(linear_start, linear_end, n_timestep, 0.1)
    elif schedule == "warmup50":
        betas = _warmup_beta(linear_start, linear_end, n_timestep, 0.5)
    elif schedule == "const":
        betas = linear_end * np.ones(n_timestep, dtype=np.float64)
    elif schedule == "jsd":  # 1/T, 1/(T-1), 1/(T-2), ..., 1
        betas = 1.0 / np.linspace(n_timestep, 1, n_timestep, dtype=np.float64)
    elif schedule == "cosine":
        timesteps = (
            torch.arange(n_timestep + 1, dtype=torch.float64) / n_timestep + cosine_s
        )
        alphas = timesteps / (1 + cosine_s) * math.pi / 2
        alphas = torch.cos(alphas).pow(2)
        alphas = alphas / alphas[0]
        betas = 1 - alphas[1:] / alphas[:-1]
        betas = betas.clamp(max=0.999)
    else:
        raise NotImplementedError(schedule)
    return betas


# gaussian diffusion trainer class


def exists(x):
    return x is not None


def default(val, d):
    if exists(val):
        return val
    return d() if isfunction(d) else d


class GaussianDiffusion(nn.Module):
    def __init__(
        self,
        denoise_fn,
        image_size,
        channels=3,
        device=None,
        loss_type="l1",
        conditional=True,
    ):
        super().__init__()
        self.channels = channels
        self.image_size = image_size
        self.denoise_fn = denoise_fn
        self.loss_type = loss_type
        self.conditional = conditional
        if self.loss_type == "l1":
            self.loss_func = nn.L1Loss(reduction="sum").to(device)
        elif self.loss_type == "l2":
            self.loss_func = nn.MSELoss(reduction="sum").to(device)

    def set_new_noise_schedule(self, device, **kwargs):
        to_torch = partial(torch.tensor, dtype=torch.float32, device=device)

        betas = make_beta_schedule(
            schedule=kwargs.get("schedule", "linear"),
            n_timestep=kwargs.get("n_timestep", 1000),
            linear_start=kwargs.get("linear_start", 1e-6),
            linear_end=kwargs.get("linear_end", 1e-2),
        )
        betas = (
            betas.detach().cpu().numpy() if isinstance(betas, torch.Tensor) else betas
        )
        alphas = 1.0 - betas
        alphas_cumprod = np.cumprod(alphas, axis=0)
        alphas_cumprod_prev = np.append(1.0, alphas_cumprod[:-1])
        self.sqrt_alphas_cumprod_prev = np.sqrt(np.append(1.0, alphas_cumprod))

        (timesteps,) = betas.shape
        self.num_timesteps = int(timesteps)
        self.register_buffer("betas", to_torch(betas))
        self.register_buffer("alphas_cumprod", to_torch(alphas_cumprod))
        self.register_buffer("alphas_cumprod_prev", to_torch(alphas_cumprod_prev))

        # calculations for diffusion q(x_t | x_{t-1}) and others
        self.register_buffer("sqrt_alphas_cumprod", to_torch(np.sqrt(alphas_cumprod)))
        self.register_buffer(
            "sqrt_one_minus_alphas_cumprod", to_torch(np.sqrt(1.0 - alphas_cumprod))
        )
        self.register_buffer(
            "log_one_minus_alphas_cumprod", to_torch(np.log(1.0 - alphas_cumprod))
        )
        self.register_buffer(
            "sqrt_recip_alphas_cumprod", to_torch(np.sqrt(1.0 / alphas_cumprod))
        )
        self.register_buffer(
            "sqrt_recipm1_alphas_cumprod", to_torch(np.sqrt(1.0 / alphas_cumprod - 1))
        )

        # calculations for posterior q(x_{t-1} | x_t, x_0)
        posterior_variance = (
            betas * (1.0 - alphas_cumprod_prev) / (1.0 - alphas_cumprod)
        )
        # above: equal to 1. / (1. / (1. - alpha_cumprod_tm1) + alpha_t / beta_t)
        self.register_buffer("posterior_variance", to_torch(posterior_variance))
        # below: log calculation clipped because the posterior variance is 0 at the beginning of the diffusion chain
        self.register_buffer(
            "posterior_log_variance_clipped",
            to_torch(np.log(np.maximum(posterior_variance, 1e-20))),
        )
        self.register_buffer(
            "posterior_mean_coef1",
            to_torch(betas * np.sqrt(alphas_cumprod_prev) / (1.0 - alphas_cumprod)),
        )
        self.register_buffer(
            "posterior_mean_coef2",
            to_torch(
                (1.0 - alphas_cumprod_prev) * np.sqrt(alphas) / (1.0 - alphas_cumprod)
            ),
        )

    def predict_start_from_noise(self, x_t, t, noise):
        return (
            self.sqrt_recip_alphas_cumprod[t] * x_t
            - self.sqrt_recipm1_alphas_cumprod[t] * noise
        )

    def q_posterior(self, x_start, x_t, t):
        posterior_mean = (
            self.posterior_mean_coef1[t] * x_start + self.posterior_mean_coef2[t] * x_t
        )
        posterior_log_variance_clipped = self.posterior_log_variance_clipped[t]
        return posterior_mean, posterior_log_variance_clipped

    def p_mean_variance(
        self, x, t, clip_denoised: bool, condition_x=None, inference=False
    ):
        batch_size = x.shape[0]
        noise_level = (
            torch.FloatTensor([self.sqrt_alphas_cumprod_prev[t + 1]])
            .repeat(batch_size, 1)
            .to(x.device)
        )
        if condition_x is not None:
            condition_x = condition_x.to(torch.float32)
        if inference:
            pred_noise = self.denoise_fn(
                torch.cat([condition_x, x], dim=1), noise_level
            )
            x_recon = self.predict_start_from_noise(x, t=t, noise=pred_noise)
        else:
            if condition_x is not None:
                x_recon = self.predict_start_from_noise(
                    x,
                    t=t,
                    noise=self.denoise_fn(
                        torch.cat([condition_x, x], dim=1), noise_level
                    ),
                )
            else:
                x_recon = self.predict_start_from_noise(
                    x, t=t, noise=self.denoise_fn(x, noise_level)
                )
        if clip_denoised:
            x_recon.clamp_(-1.0, 1.0)

        model_mean, posterior_log_variance = self.q_posterior(
            x_start=x_recon, x_t=x, t=t
        )
        if inference:
            return model_mean, posterior_log_variance, x_recon, pred_noise
        else:
            return model_mean, posterior_log_variance

    @torch.no_grad()
    def p_sample(self, x, t, clip_denoised=True, condition_x=None, inference=False):
        if inference:
            model_mean, model_log_variance, x_recon, pred_noise = self.p_mean_variance(
                x=x,
                t=t,
                clip_denoised=clip_denoised,
                condition_x=condition_x,
                inference=inference,
            )
            noise = torch.randn_like(x) if t > 0 else torch.zeros_like(x)
            return (
                model_mean + noise * (0.5 * model_log_variance).exp(),
                x_recon,
                pred_noise,
            )
        else:
            model_mean, model_log_variance = self.p_mean_variance(
                x=x,
                t=t,
                clip_denoised=clip_denoised,
                condition_x=condition_x,
                inference=inference,
            )
            noise = torch.randn_like(x) if t > 0 else torch.zeros_like(x)
            return model_mean + noise * (0.5 * model_log_variance).exp()

    @torch.no_grad()
    def p_sample_loop(self, x_in, continous=False):
        device = self.betas.device
        sample_inter = 1 | (self.num_timesteps // 10)
        if not self.conditional:
            shape = x_in
            img = torch.randn(shape, device=device)
            ret_img = img
            for i in reversed(range(0, self.num_timesteps)):
                img = self.p_sample(img, i)
                if i % sample_inter == 0:
                    ret_img = torch.cat([ret_img, img], dim=0)
        else:
            x = x_in
            shape = x.shape
            img = torch.randn(shape, device=device)
            ret_img = x
            for i in reversed(range(0, self.num_timesteps)):
                img = self.p_sample(img, i, condition_x=x)
                if i % sample_inter == 0:
                    ret_img = torch.cat([ret_img, img], dim=0)
        if continous:
            return ret_img
        else:
            return ret_img[-shape[0] :]

    @torch.no_grad()
    def ddim_sample(
        self, x_in, sampling_timesteps, ddim_sampling_eta, continuous=False
    ):
        shape = x_in.shape
        batch = shape[0]
        device = self.betas.device
        total_timesteps = self.num_timesteps
        eta = ddim_sampling_eta

        times = torch.linspace(
            -1, total_timesteps - 1, steps=sampling_timesteps + 1
        ).int()
        times = torch.flip(times, dims=[0])
        time_pairs = zip(times[:-1], times[1:])
        img = torch.randn(
            (batch, self.channels, self.image_size, self.image_size), device=device
        )
        imgs = [img]
        x_start = None
        for time, time_next in time_pairs:
            _, x_start, pred_noise = self.p_sample(
                img, time, condition_x=x_in, inference=True
            )
            imgs.append(img)
            if time_next < 0:
                img = x_start
                continue
            alpha = self.alphas_cumprod[time]
            alpha_next = self.alphas_cumprod[time_next]
            sigma = (
                eta * ((1 - alpha / alpha_next) * (1 - alpha_next) / (1 - alpha)).sqrt()
            )
            c = (1 - alpha_next - sigma**2).sqrt()
            noise = torch.randn_like(img, device=device)
            img = x_start * alpha_next.sqrt() + c * pred_noise + sigma * noise
        ret = img if not continuous else torch.stack(imgs, dim=1)
        return ret

    @torch.no_grad()
    def sample(self, batch_size=1, continous=False):
        image_size = self.image_size
        channels = self.channels
        return self.p_sample_loop(
            (batch_size, channels, image_size, image_size), continous
        )

    @torch.no_grad()
    def super_resolution(self, x_in, continous, **kwargs):
        if kwargs.get("sampling", "ddpm") == "ddim":
            sampling_timesteps = kwargs.get("sampling_timesteps")
            ddim_sampling_eta = kwargs.get("ddim_sampling_eta")
            return self.ddim_sample(
                x_in, sampling_timesteps, ddim_sampling_eta, continous
            )
        else:
            return self.p_sample_loop(x_in, continous)

    def q_sample(self, x_start, continuous_sqrt_alpha_cumprod, noise=None):
        noise = default(noise, lambda: torch.randn_like(x_start))

        # random gama
        return (
            continuous_sqrt_alpha_cumprod * x_start
            + (1 - continuous_sqrt_alpha_cumprod**2).sqrt() * noise
        )

    def p_losses(self, x_in, noise=None):
        x_start = x_in["HR"]
        [b, c, h, w] = x_start.shape
        t = np.random.randint(1, self.num_timesteps + 1)
        continuous_sqrt_alpha_cumprod = torch.FloatTensor(
            np.random.uniform(
                self.sqrt_alphas_cumprod_prev[t - 1],
                self.sqrt_alphas_cumprod_prev[t],
                size=b,
            )
        ).to(x_start.device)
        continuous_sqrt_alpha_cumprod = continuous_sqrt_alpha_cumprod.view(b, -1)

        noise = default(noise, lambda: torch.randn_like(x_start))
        x_noisy = self.q_sample(
            x_start=x_start,
            continuous_sqrt_alpha_cumprod=continuous_sqrt_alpha_cumprod.view(
                -1, 1, 1, 1
            ),
            noise=noise,
        )

        if not self.conditional:
            x_recon = self.denoise_fn(x_noisy, continuous_sqrt_alpha_cumprod)
        else:
            x_recon = self.denoise_fn(
                torch.cat([x_in["SR"], x_noisy], dim=1), continuous_sqrt_alpha_cumprod
            )

        loss = self.loss_func(noise, x_recon)
        return loss

    def forward(self, sr, hr, *args, **kwargs):
        return self.p_losses({"HR": hr, "SR": sr}, *args, **kwargs)


def exists(x):
    return x is not None


def default(val, d):
    if exists(val):
        return val
    return d() if isfunction(d) else d


# PositionalEncoding Source： https://github.com/lmnt-com/wavegrad/blob/master/src/wavegrad/model.py
class PositionalEncoding(nn.Module):
    def __init__(self, dim):
        super().__init__()
        self.dim = dim

    def forward(self, noise_level):
        count = self.dim // 2
        step = (
            torch.arange(count, dtype=noise_level.dtype, device=noise_level.device)
            / count
        )
        encoding = noise_level.unsqueeze(1) * torch.exp(
            -math.log(1e4) * step.unsqueeze(0)
        )
        encoding = torch.cat([torch.sin(encoding), torch.cos(encoding)], dim=-1)
        return encoding


class FeatureWiseAffine(nn.Module):
    def __init__(self, in_channels, out_channels, use_affine_level=False):
        super(FeatureWiseAffine, self).__init__()
        self.use_affine_level = use_affine_level
        self.noise_func = nn.Sequential(
            nn.Linear(in_channels, out_channels * (1 + self.use_affine_level))
        )

    def forward(self, x, noise_embed):
        batch = x.shape[0]
        if self.use_affine_level:
            gamma, beta = (
                self.noise_func(noise_embed).view(batch, -1, 1, 1).chunk(2, dim=1)
            )
            x = (1 + gamma) * x + beta
        else:
            x = x + self.noise_func(noise_embed).view(batch, -1, 1, 1)
        return x


class Swish(nn.Module):
    def forward(self, x):
        return x * torch.sigmoid(x)


class Upsample(nn.Module):
    def __init__(self, dim):
        super().__init__()
        self.up = nn.Upsample(scale_factor=2, mode="nearest")
        self.conv = nn.Conv2d(dim, dim, 3, padding=1)

    def forward(self, x):
        return self.conv(self.up(x))


class Downsample(nn.Module):
    def __init__(self, dim):
        super().__init__()
        self.conv = nn.Conv2d(dim, dim, 3, 2, 1)

    def forward(self, x):
        return self.conv(x)


# building block modules


class Block(nn.Module):
    def __init__(self, dim, dim_out, groups=32, dropout=0):
        super().__init__()
        self.block = nn.Sequential(
            nn.GroupNorm(groups, dim),
            Swish(),
            nn.Dropout(dropout) if dropout != 0 else nn.Identity(),
            nn.Conv2d(dim, dim_out, 3, padding=1),
        )

    def forward(self, x):
        return self.block(x)


class ResnetBlock(nn.Module):
    def __init__(
        self,
        dim,
        dim_out,
        noise_level_emb_dim=None,
        dropout=0,
        use_affine_level=False,
        norm_groups=32,
    ):
        super().__init__()
        self.noise_func = FeatureWiseAffine(
            noise_level_emb_dim, dim_out, use_affine_level
        )

        self.block1 = Block(dim, dim_out, groups=norm_groups)
        self.block2 = Block(dim_out, dim_out, groups=norm_groups, dropout=dropout)
        self.res_conv = nn.Conv2d(dim, dim_out, 1) if dim != dim_out else nn.Identity()

    def forward(self, x, time_emb):
        b, c, h, w = x.shape
        h = self.block1(x)
        h = self.noise_func(h, time_emb)
        h = self.block2(h)
        return h + self.res_conv(x)


class SelfAttention(nn.Module):
    def __init__(self, in_channel, n_head=1, norm_groups=32):
        super().__init__()

        self.n_head = n_head

        self.norm = nn.GroupNorm(norm_groups, in_channel)
        self.qkv = nn.Conv2d(in_channel, in_channel * 3, 1, bias=False)
        self.out = nn.Conv2d(in_channel, in_channel, 1)

    def forward(self, input):
        batch, channel, height, width = input.shape
        n_head = self.n_head
        head_dim = channel // n_head

        norm = self.norm(input)
        qkv = self.qkv(norm).view(batch, n_head, head_dim * 3, height, width)
        query, key, value = qkv.chunk(3, dim=2)  # bhdyx

        attn = torch.einsum(
            "bnchw, bncyx -> bnhwyx", query, key
        ).contiguous() / math.sqrt(channel)
        attn = attn.view(batch, n_head, height, width, -1)
        attn = torch.softmax(attn, -1)
        attn = attn.view(batch, n_head, height, width, height, width)

        out = torch.einsum("bnhwyx, bncyx -> bnchw", attn, value).contiguous()
        out = self.out(out.view(batch, channel, height, width))

        return out + input


class ResnetBlocWithAttn(nn.Module):
    def __init__(
        self,
        dim,
        dim_out,
        *,
        noise_level_emb_dim=None,
        norm_groups=32,
        dropout=0,
        with_attn=False
    ):
        super().__init__()
        self.with_attn = with_attn
        self.res_block = ResnetBlock(
            dim, dim_out, noise_level_emb_dim, norm_groups=norm_groups, dropout=dropout
        )
        if with_attn:
            self.attn = SelfAttention(dim_out, norm_groups=norm_groups)

    def forward(self, x, time_emb):
        x = self.res_block(x, time_emb)
        if self.with_attn:
            x = self.attn(x)
        return x


class UNet(nn.Module):
    def __init__(
        self,
        in_channel=6,
        out_channel=3,
        image_size=256,
        with_noise_level_emb=True,
        **kwargs
    ):
        super().__init__()

        inner_channel = kwargs.get("inner_channel", 64)
        norm_groups = kwargs.get("norm_groups", 32)
        channel_mults = kwargs.get("channel_mults", [1, 2, 4, 4, 8, 8])
        attn_res = [kwargs.get("attn_res", 16)]
        res_blocks = kwargs.get("res_blocks", 3)
        dropout = kwargs.get("dropout", 0)

        if with_noise_level_emb:
            noise_level_channel = inner_channel
            self.noise_level_mlp = nn.Sequential(
                PositionalEncoding(inner_channel),
                nn.Linear(inner_channel, inner_channel * 4),
                Swish(),
                nn.Linear(inner_channel * 4, inner_channel),
            )
        else:
            noise_level_channel = None
            self.noise_level_mlp = None

        num_mults = len(channel_mults)
        pre_channel = inner_channel
        feat_channels = [pre_channel]
        now_res = image_size
        downs = [nn.Conv2d(in_channel, inner_channel, kernel_size=3, padding=1)]
        for ind in range(num_mults):
            is_last = ind == num_mults - 1
            use_attn = now_res in attn_res
            channel_mult = inner_channel * channel_mults[ind]
            for _ in range(0, res_blocks):
                downs.append(
                    ResnetBlocWithAttn(
                        pre_channel,
                        channel_mult,
                        noise_level_emb_dim=noise_level_channel,
                        norm_groups=norm_groups,
                        dropout=dropout,
                        with_attn=use_attn,
                    )
                )
                feat_channels.append(channel_mult)
                pre_channel = channel_mult
            if not is_last:
                downs.append(Downsample(pre_channel))
                feat_channels.append(pre_channel)
                now_res = now_res // 2
        self.downs = nn.ModuleList(downs)

        self.mid = nn.ModuleList(
            [
                ResnetBlocWithAttn(
                    pre_channel,
                    pre_channel,
                    noise_level_emb_dim=noise_level_channel,
                    norm_groups=norm_groups,
                    dropout=dropout,
                    with_attn=True,
                ),
                ResnetBlocWithAttn(
                    pre_channel,
                    pre_channel,
                    noise_level_emb_dim=noise_level_channel,
                    norm_groups=norm_groups,
                    dropout=dropout,
                    with_attn=False,
                ),
            ]
        )

        ups = []
        for ind in reversed(range(num_mults)):
            is_last = ind < 1
            use_attn = now_res in attn_res
            channel_mult = inner_channel * channel_mults[ind]
            for _ in range(0, res_blocks + 1):
                ups.append(
                    ResnetBlocWithAttn(
                        pre_channel + feat_channels.pop(),
                        channel_mult,
                        noise_level_emb_dim=noise_level_channel,
                        norm_groups=norm_groups,
                        dropout=dropout,
                        with_attn=use_attn,
                    )
                )
                pre_channel = channel_mult
            if not is_last:
                ups.append(Upsample(pre_channel))
                now_res = now_res * 2

        self.ups = nn.ModuleList(ups)

        self.final_conv = Block(
            pre_channel, default(out_channel, in_channel), groups=norm_groups
        )

    def forward(self, x, time):
        t = self.noise_level_mlp(time) if exists(self.noise_level_mlp) else None

        feats = []
        for layer in self.downs:
            if isinstance(layer, ResnetBlocWithAttn):
                x = layer(x, t)
            else:
                x = layer(x)
            feats.append(x)

        for layer in self.mid:
            if isinstance(layer, ResnetBlocWithAttn):
                x = layer(x, t)
            else:
                x = layer(x)

        for layer in self.ups:
            if isinstance(layer, ResnetBlocWithAttn):
                x = layer(torch.cat((x, feats.pop()), dim=1), t)
            else:
                x = layer(x)

        return self.final_conv(x)


import torch.nn as nn
from torch.nn import init

####################
# initialize
####################


def weights_init_normal(m, std=0.02):
    classname = m.__class__.__name__
    if classname.find("Conv") != -1:
        init.normal_(m.weight.data, 0.0, std)
        if m.bias is not None:
            m.bias.data.zero_()
    elif classname.find("Linear") != -1:
        init.normal_(m.weight.data, 0.0, std)
        if m.bias is not None:
            m.bias.data.zero_()
    elif classname.find("BatchNorm2d") != -1:
        init.normal_(m.weight.data, 1.0, std)  # BN also uses norm
        init.constant_(m.bias.data, 0.0)


def weights_init_kaiming(m, scale=1):
    classname = m.__class__.__name__
    if classname.find("Conv2d") != -1:
        init.kaiming_normal_(m.weight.data, a=0, mode="fan_in")
        m.weight.data *= scale
        if m.bias is not None:
            m.bias.data.zero_()
    elif classname.find("Linear") != -1:
        init.kaiming_normal_(m.weight.data, a=0, mode="fan_in")
        m.weight.data *= scale
        if m.bias is not None:
            m.bias.data.zero_()
    elif classname.find("BatchNorm2d") != -1:
        init.constant_(m.weight.data, 1.0)
        init.constant_(m.bias.data, 0.0)


def weights_init_orthogonal(m):
    classname = m.__class__.__name__
    if classname.find("Conv") != -1:
        init.orthogonal_(m.weight.data, gain=1)
        if m.bias is not None:
            m.bias.data.zero_()
    elif classname.find("Linear") != -1:
        init.orthogonal_(m.weight.data, gain=1)
        if m.bias is not None:
            m.bias.data.zero_()
    elif classname.find("BatchNorm2d") != -1:
        init.constant_(m.weight.data, 1.0)
        init.constant_(m.bias.data, 0.0)


def init_weights(net, init_type="kaiming", scale=1, std=0.02):
    # scale for 'kaiming', std for 'normal'
    if init_type == "normal":
        weights_init_normal_ = functools.partial(weights_init_normal, std=std)
        net.apply(weights_init_normal_)
    elif init_type == "kaiming":
        weights_init_kaiming_ = functools.partial(weights_init_kaiming, scale=scale)
        net.apply(weights_init_kaiming_)
    elif init_type == "orthogonal":
        net.apply(weights_init_orthogonal)
    else:
        raise NotImplementedError(
            "initialization method [{:s}] not implemented".format(init_type)
        )


class l1Loss(nn.Module):
    def __init__(self, device):
        super(l1Loss, self).__init__()

    def forward(self, l_pix, target):
        b, c, h, w = target.shape
        l_pix = l_pix.sum() / int(b * c * h * w)
        return l_pix



import math
from functools import reduce, lru_cache

import pandas as pd
import numpy as np

from arcgis.geometry._types import Geometry
from arcgis.features.geo._accessor import GeoAccessor
from arcgis.features.geo._accessor import GeoSeriesAccessor
from arcgis.features.geo._accessor import _is_geoenabled
from arcgis.features.geo._array import GeoArray

_HASARCPY, _HASSHAPELY = None, None


# ----------------------------------------------------------------------
@lru_cache(maxsize=10)
def _check_geometry_engine():
    """checks if the geometry engine exists"""
    global _HASARCPY
    global _HASSHAPELY
    if _HASARCPY is None:
        try:
            import arcpy

            _HASARCPY = True
        except:
            _HASARCPY = False
    if _HASSHAPELY is None:
        try:
            import shapely

            _HASSHAPELY = True
        except:
            _HASSHAPELY = False
    return _HASARCPY, _HASSHAPELY


# --------------------------------------------------------------------------
def _overlay_difference(df1, df2):
    """
    Overlay Difference operation used in overlay function
    """
    # Spatial Index to create intersections
    spatial_index = df2.spatial.sindex()
    bbox = df1[df1.spatial.name].geom.extent
    fn = lambda x: spatial_index.intersect(x)
    sidx = bbox.apply(fn)
    # Create differences
    new_g = []
    for geom, neighbours in zip(df1[df1.spatial.name], sidx):
        new = reduce(
            lambda x, y: x.difference(y).buffer(0),
            [geom] + list(df2[df2.spatial.name].iloc[neighbours]),
        )
        new_g.append(new)
    differences = pd.Series(GeoArray(new_g), index=df1.index)
    q = differences.isnull()
    geom_diff = differences[~q].copy()
    dfdiff = df1[~q].copy()
    dfdiff[dfdiff.spatial.name] = geom_diff
    return dfdiff


# --------------------------------------------------------------------------
def _symmetric_difference(df1, df2):
    """
    The symmetric difference, also known as the disjunctive union, of two
    sets is the set of elements which are in either of the sets and not in
    their intersection.

    :returns: pd.DataFrame (Spatially enabled DataFrame)

    """
    dfdiff1 = _overlay_difference(df1, df2)
    dfdiff2 = _overlay_difference(df2, df1)
    dfdiff1["__idx1"] = range(len(dfdiff1))
    dfdiff2["__idx2"] = range(len(dfdiff2))
    dfdiff1["__idx2"] = np.nan
    dfdiff2["__idx1"] = np.nan
    dfsym = dfdiff1.merge(
        dfdiff2, on=["__idx1", "__idx2"], how="outer", suffixes=["_1", "_2"]
    )
    geometry = dfsym.SHAPE_1.copy()
    geometry[dfsym.SHAPE_1.isnull()] = dfsym.loc[dfsym.SHAPE_1.isnull(), "SHAPE_2"]
    dfsym.drop(["SHAPE_1", "SHAPE_2"], axis=1, inplace=True)
    dfsym.reset_index(drop=True, inplace=True)
    dfsym = pd.DataFrame(data=dfsym)
    dfsym["SHAPE"] = geometry
    dfsym.spatial.name
    return dfsym


# --------------------------------------------------------------------------
def _overlay_intersection(df1, df2):
    """
    Overlay Intersection operation used in overlay function
    """

    def _ops_intersection(this, other):
        """performs the series vs series intersection"""
        lu = {"polygon": 4, "point": 1, "multipoint": 1, "line": 2, "polyline": 2}
        null_value = None
        gtype = lu[this.geom.geometry_type[0]]
        this, other = this.align(other)
        data = np.array(
            [
                getattr(this_elem, "intersect")(other_elem, gtype)
                if not this_elem.is_empty | other_elem.is_empty
                else null_value
                for this_elem, other_elem in zip(this, other)
            ]
        )

        return pd.Series(GeoArray(data), index=this.index)

    # Spatial Index to create intersections
    spatial_index = df2.spatial.sindex()
    bbox = df1[df1.spatial.name].geom.extent
    fn = lambda x: spatial_index.intersect(x)
    sidx = bbox.apply(fn)
    # Create pairs of geometries in both dataframes to be intersected
    nei = []
    for i, j in enumerate(sidx):
        for k in j:
            nei.append([i, k])
    if nei != []:
        pairs = pd.DataFrame(nei, columns=["__idx1", "__idx2"])
        left = df1[df1.spatial.name].take(pairs["__idx1"].values)
        left.reset_index(drop=True, inplace=True)
        right = df2[df2.spatial.name].take(pairs["__idx2"].values)
        right.reset_index(drop=True, inplace=True)
        intersections = _ops_intersection(left, right).geom.buffer(0)
        intersections.name = "intersections"

        # only keep actual intersecting geometries
        pairs_intersect = pairs[~intersections.isnull()]
        geom_intersect = intersections[~intersections.isnull()]

        # merge data for intersecting geometries
        df1 = df1.reset_index(drop=True)
        df2 = df2.reset_index(drop=True)
        dfinter = pairs_intersect.merge(
            df1.drop(df1.spatial.name, axis=1), left_on="__idx1", right_index=True
        )
        dfinter = dfinter.merge(
            df2.drop(df2.spatial.name, axis=1),
            left_on="__idx2",
            right_index=True,
            suffixes=["_1", "_2"],
        )
        dfinter["SHAPE"] = geom_intersect
        return dfinter
    else:
        return pd.DataFrame(
            data=[],
            columns=list(set(df1.columns).union(df2.columns)) + ["__idx1", "__idx2"],
        )


def _overlay_intersection2(df1, df2):
    """
    Overlay Intersection operation used in overlay function
    """

    def _ops_intersection(this, other):
        """performs the series vs series intersection"""
        lu = {"polygon": 4, "point": 1, "multipoint": 1, "line": 2, "polyline": 2}
        null_value = None
        gtype = lu[this.geom.geometry_type[0]]
        this, other = this.align(other)
        data = np.array(
            [
                getattr(this_elem, "intersect")(other_elem, gtype)
                if not this_elem.is_empty | other_elem.is_empty
                else null_value
                for this_elem, other_elem in zip(this, other)
            ]
        )
        return pd.Series(GeoArray(data), index=this.index)

    # Spatial Index to create intersections
    spatial_index = df2.spatial.sindex()
    bbox = df1[df1.spatial.name].geom.extent
    fn = lambda x: spatial_index.intersect(x)
    sidx = bbox.apply(fn)
    # Create pairs of geometries in both dataframes to be intersected
    nei = []
    for i, j in enumerate(sidx):
        for k in j:
            nei.append([i, k])
    if nei != []:
        pairs = pd.DataFrame(nei, columns=["__idx1", "__idx2"])
        left = df1[df1.spatial.name].values.take(pairs["__idx1"].values)
        left = pd.Series(left)
        left.reset_index(drop=True, inplace=True)
        right = df2[df2.spatial.name].values.take(pairs["__idx2"].values)
        right.reset_index(drop=True, inplace=True)
        intersections = _ops_intersection(left, right).geom.buffer(0)
        intersections.name = "intersections"
        # only keep actual intersecting geometries
        pairs_intersect = pairs[~intersections.isnull()]
        geom_intersect = intersections[~intersections.isnull()]
        # merge data for intersecting geometries
        df1 = df1.reset_index(drop=True)
        df2 = df2.reset_index(drop=True)
        dfinter = pairs_intersect.merge(
            df1.drop(df1.spatial.name, axis=1), left_on="__idx1", right_index=True
        )
        dfinter = dfinter.merge(
            df2.drop(df2.spatial.name, axis=1),
            left_on="__idx2",
            right_index=True,
            suffixes=["_1", "_2"],
        )
        dfinter["SHAPE"] = geom_intersect
        return dfinter
    else:
        return pd.DataFrame(
            data=[],
            columns=list(set(df1.columns).union(df2.columns)) + ["__idx1", "__idx2"],
        )


# --------------------------------------------------------------------------
def _erase(df1, df2):
    """
    This overlay operation erases geometries by the second SeDF object.

    """
    # Spatial Index to create intersections
    spatial_index = df2.sindex
    bbox = df1.geometry.apply(lambda x: x.bounds)
    sidx = bbox.apply(lambda x: list(spatial_index.intersection(x)))
    # Create differences
    new_g = []
    for geom, neighbours in zip(df1.geometry, sidx):
        new = reduce(
            lambda x, y: x.difference(y).buffer(0),
            [geom] + list(df2.geometry.iloc[neighbours]),
        )
        new_g.append(new)
    differences = pd.Series(new_g, index=df1.index)
    geom_diff = differences[~differences.is_empty].copy()
    dfdiff = df1[~differences.is_empty].copy()
    dfdiff[dfdiff._geometry_column_name] = geom_diff
    return dfdiff


def _overlay_union(df1, df2):
    """
    Overlay Union operation used in overlay function
    """
    dfinter = _overlay_intersection(df1, df2)
    dfsym = _symmetric_difference(df1, df2)
    dfunion = pd.concat([dfinter, dfsym], ignore_index=True, sort=False)
    columns = list(dfunion.columns)
    return dfunion.reindex(columns=columns)


def overlay_dask(sdf1, sdf2, op="union"):
    """
    Perform spatial overlay operations between two polygons Spatially
    Enabled DataFrames.

    =========================    =========================================================
    **Parameter**                 **Description**
    -------------------------    ---------------------------------------------------------
    sdf1                         Required Spatially Enabled DataFrame. The dataframe to have the operation performed on.
    -------------------------    ---------------------------------------------------------
    sdf2                         Required Spatially Enabled DataFrame. The geometry to perform the operation from.
    -------------------------    ---------------------------------------------------------
    op                           Optional String. The spatial operation to perform.  The
                                 allowed value are: union, erase, identity, intersection
    =========================    =========================================================

    :returns: Spatially enabled DataFrame (pd.DataFrame)

    """
    allowed_hows = [
        "intersection",
        "union",
        "identity",
        "symmetric_difference",
        "difference",
        "erase",
    ]

    op = str(op).lower()

    _hasao, _hasshp = _check_geometry_engine()

    if (
        _hasao == False
        and _hasshp
        and sdf1.spatial.geometry_type != ["polygon"]
        and sdf2.spatial.geometry_type != ["polygon"]
    ):
        raise ValueError(
            ("Using shapely's geometry engine only " "support Polygon geometries.")
        )

    if (_hasao or _hasshp) and op in allowed_hows:
        if op in ["union", "identity"]:
            return _overlay_union(sdf1, sdf2)
        elif op in ["difference", "erase"]:
            return _overlay_difference(sdf1, sdf2)
        elif op == "intersection":
            return _overlay_intersection(sdf1, sdf2)
        elif op == "symmetric_difference":
            return _symmetric_difference(sdf1, sdf2)
    return None



# Copyright 2018-2019 Open-MMLab.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# model settings
model = dict(
    type="RPN",
    pretrained="open-mmlab://detectron2/resnet50_caffe",
    backbone=dict(
        type="ResNet",
        depth=50,
        num_stages=3,
        strides=(1, 2, 2),
        dilations=(1, 1, 1),
        out_indices=(2,),
        frozen_stages=1,
        norm_cfg=dict(type="BN", requires_grad=False),
        norm_eval=True,
        style="caffe",
    ),
    neck=None,
    rpn_head=dict(
        type="RPNHead",
        in_channels=1024,
        feat_channels=1024,
        anchor_generator=dict(
            type="AnchorGenerator",
            scales=[2, 4, 8, 16, 32],
            ratios=[0.5, 1.0, 2.0],
            strides=[16],
        ),
        bbox_coder=dict(
            type="DeltaXYWHBBoxCoder",
            target_means=[0.0, 0.0, 0.0, 0.0],
            target_stds=[1.0, 1.0, 1.0, 1.0],
        ),
        loss_cls=dict(type="CrossEntropyLoss", use_sigmoid=True, loss_weight=1.0),
        loss_bbox=dict(type="L1Loss", loss_weight=1.0),
    ),
    # model training and testing settings
    train_cfg=dict(
        rpn=dict(
            assigner=dict(
                type="MaxIoUAssigner",
                pos_iou_thr=0.7,
                neg_iou_thr=0.3,
                min_pos_iou=0.3,
                ignore_iof_thr=-1,
            ),
            sampler=dict(
                type="RandomSampler",
                num=256,
                pos_fraction=0.5,
                neg_pos_ub=-1,
                add_gt_as_proposals=False,
            ),
            allowed_border=0,
            pos_weight=-1,
            debug=False,
        )
    ),
    test_cfg=dict(
        rpn=dict(
            nms_across_levels=False,
            nms_pre=12000,
            nms_post=2000,
            max_num=2000,
            nms_thr=0.7,
            min_bbox_size=0,
        )
    ),
)



from __future__ import annotations
from arcgis.auth.tools import LazyLoader
import base64
import hashlib
import logging
import socket
import struct
import typing
from urllib.parse import urlparse
import requests
from requests.auth import AuthBase
from requests.exceptions import HTTPError
from ._utils import _split_username

try:
    pywintypes = LazyLoader("pywintypes", strict=True)
    sspi = LazyLoader("sspi", strict=True)
    sspicon = LazyLoader("sspicon", strict=True)
    win32security = LazyLoader("win32security", strict=True)

    HAS_GSSAPI = True
except ImportError:
    HAS_GSSAPI = False
from ._utils import parse_url, assemble_url

from ._schain import SupportMultiAuth

_logger = logging.getLogger(__name__)

__all__ = ["EsriHttpNegotiateAuth"]


###########################################################################
class EsriHttpNegotiateAuth(AuthBase, SupportMultiAuth):
    """
    This class is dervived from the `requests_negotiate_sspi` package. It
    extends the `HttpNegotiateAuth` class inorder to handle the Esri security
    model.

    ================    ===============================================================
    **Parameter**       **Description**
    ----------------    ---------------------------------------------------------------
    username            Optional String. The username account with the domain (DOMAIN\\USERNAME).
    ----------------    ---------------------------------------------------------------
    password            Optional String. The username's password.
    ----------------    ---------------------------------------------------------------
    service             Optional String. Kerberos Service type for remote Service Principal Name.
    ----------------    ---------------------------------------------------------------
    host                Optional String. Host name for Service Principal Name.
    ----------------    ---------------------------------------------------------------
    delegate            Optional Boolean.  Indicates that the user's credentials are to be delegated to the server.
    ----------------    ---------------------------------------------------------------
    referer             Optional String. The referer for the Esri token.
    ----------------    ---------------------------------------------------------------
    verify_cert         Optional Boolean. When false, certificate errors are ignored. The default is True.
    ================    ===============================================================

    If username and password are not specified, the user's default credentials are used.
    This allows for single-sign-on to domain resources if the user is currently logged on
    with a domain account.

    """

    _auth_info = None
    _service = "HTTP"
    _host = None
    _delegate = False

    # ----------------------------------------------------------------------
    def __init__(
        self,
        username=None,
        password=None,
        service=None,
        host=None,
        delegate=False,
        *,
        referer: str = "http",
        verify_cert: bool = True,
        **kwargs,
    ):
        """Create a new Negotiate auth handler

        Args:
         username: Username.
         password: Password.
         domain: NT Domain name.
             Default: '.' for local account.
         service: Kerberos Service type for remote Service Principal Name.
             Default: 'HTTP'
         host: Host name for Service Principal Name.
             Default: Extracted from request URI
         delegate: Indicates that the user's credentials are to be delegated to the server.
             Default: False

         If username and password are not specified, the user's default credentials are used.
         This allows for single-sign-on to domain resources if the user is currently logged on
         with a domain account.
        """
        domain = None
        try:
            username, domain = _split_username(username)
        except:
            username, domain = username, None
        if HAS_GSSAPI == False:
            raise Exception(
                "The system does not have the required dependencies"
                " or is not a Windows based operating system."
            )
        if domain is None:
            domain = "."

        if username is not None and password is not None:
            self._auth_info = (username, domain, password)

        if service is not None:
            self._service = service

        if host is not None:
            self._host = host

        self._delegate = delegate
        self._server_log: dict[str, typing.Any] = {}
        self._referer: str = referer
        self._verify_cert: bool = verify_cert
        self._proxy: dict[str, typing.Any] = kwargs.pop("proxy", None)

    # ----------------------------------------------------------------------
    def __str__(self):
        return f"<{self.__class__.__name__}>"

    # ----------------------------------------------------------------------
    def __repr__(self):
        return f"<{self.__class__.__name__}>"

    # ----------------------------------------------------------------------
    def generate_token(self, r, scheme, args):
        """
        Generates the Server Token

        Args:
         r: the requests.Response object
         schema: string. Http or HTTPS
         args: a list of optional arguments for the requests' send method.

        """
        parsed = parse_url(url=r.url)
        server_url = assemble_url(parsed)
        token_url: str = None
        if server_url in self._server_log:
            token_url: str = self._server_log[server_url]
        elif r.text.lower().find("token required") > -1:
            resp = requests.get(
                f"{server_url}/rest/info",
                params={"f": "json"},
                auth=self,
                headers={"referer": self._referer},
                verify=self._verify_cert,
                proxies=self._proxy,
            ).json()
            self._server_log[parsed.netloc] = resp["authInfo"]["tokenServicesUrl"]
            token_url: str = self._server_log[parsed.netloc]

        if token_url:
            r.content
            r.raw.release_conn()
            request = r.request.copy()

            # this is important for some web applications that store
            # authentication-related info in cookies
            if r.headers.get("set-cookie"):
                request.headers["Cookie"] = r.headers.get("set-cookie")
            postdata = {
                "request": "getToken",
                "serverURL": server_url,
                "referer": "http",
                "f": "json",
            }
            resp = requests.post(
                token_url,
                params=postdata,
                auth=self,
                headers={"referer": self._referer},
                verify=self._verify_cert,
                proxies=self._proxy,
            ).json()
            token_str = resp["token"]
            request.headers["X-Esri-Authorization"] = f"Bearer {token_str}"

            return request.copy()
        else:
            return r

    # ----------------------------------------------------------------------
    def _retry_using_http_Negotiate_auth(self, response, scheme, args):
        """performs the NTLM authorization"""
        if "Authorization" in response.request.headers:
            return response

        if self._host is None:
            targeturl = urlparse(response.request.url)
            self._host = targeturl.hostname
            try:
                self._host = socket.getaddrinfo(
                    self._host, None, 0, 0, 0, socket.AI_CANONNAME
                )[0][3]
            except socket.gaierror as e:
                _logger.info(
                    "Skipping canonicalization of name %s due to error: %s",
                    self._host,
                    e,
                )

        targetspn = "{}/{}".format(self._service, self._host)

        # We request mutual auth by default
        scflags = sspicon.ISC_REQ_MUTUAL_AUTH

        if self._delegate:
            scflags |= sspicon.ISC_REQ_DELEGATE

        # Set up SSPI connection structure
        pkg_info = win32security.QuerySecurityPackageInfo(scheme)
        clientauth = sspi.ClientAuth(
            scheme,
            targetspn=targetspn,
            auth_info=self._auth_info,
            scflags=scflags,
            datarep=sspicon.SECURITY_NETWORK_DREP,
        )
        sec_buffer = win32security.PySecBufferDescType()

        # Channel Binding Hash (aka Extended Protection for Authentication)
        # If this is a SSL connection, we need to hash the peer certificate, prepend the RFC5929 channel binding type,
        # and stuff it into a SEC_CHANNEL_BINDINGS structure.
        # This should be sent along in the initial handshake or Kerberos auth will fail.
        if hasattr(response, "peercert") and response.peercert is not None:
            md = hashlib.sha256()
            md.update(response.peercert)
            appdata = "tls-server-end-point:".encode("ASCII") + md.digest()
            cbtbuf = win32security.PySecBufferType(
                pkg_info["MaxToken"], sspicon.SECBUFFER_CHANNEL_BINDINGS
            )
            cbtbuf.Buffer = struct.pack(
                "LLLLLLLL{}s".format(len(appdata)),
                0,
                0,
                0,
                0,
                0,
                0,
                len(appdata),
                32,
                appdata,
            )
            sec_buffer.append(cbtbuf)

        content_length = int(
            response.request.headers.get("Content-Length", "0"), base=10
        )

        if hasattr(response.request.body, "seek"):
            if content_length > 0:
                response.request.body.seek(-content_length, 1)
            else:
                response.request.body.seek(0, 0)

        # Consume content and release the original connection
        # to allow our new request to reuse the same one.
        response.content
        response.raw.release_conn()
        request = response.request.copy()

        # this is important for some web applications that store
        # authentication-related info in cookies
        if response.headers.get("set-cookie"):
            request.headers["Cookie"] = response.headers.get("set-cookie")

        # Send initial challenge auth header
        try:
            error, auth = clientauth.authorize(sec_buffer)
            request.headers["Authorization"] = "{} {}".format(
                scheme, base64.b64encode(auth[0].Buffer).decode("ASCII")
            )
            _logger.debug(
                "Sending Initial Context Token - error={} authenticated={}".format(
                    error, clientauth.authenticated
                )
            )
        except pywintypes.error as e:
            _logger.debug("Error calling {}: {}".format(e[1], e[2]), exc_info=e)
            return response

        # A streaming response breaks authentication.
        # This can be fixed by not streaming this request, which is safe
        # because the returned response3 will still have stream=True set if
        # specified in args. In addition, we expect this request to give us a
        # challenge and not the real content, so the content will be short
        # anyway.
        args_nostream = dict(args, stream=False)
        response2 = response.connection.send(request, **args_nostream)

        # Should get another 401 if we are doing challenge-response (NTLM)
        if response2.status_code != 401:
            # Kerberos may have succeeded; if so, finalize our auth context
            final = response2.headers.get("WWW-Authenticate")
            if final is not None:
                try:
                    # Sometimes Windows seems to forget to prepend 'Negotiate' to the success response,
                    # and we get just a bare chunk of base64 token. Not sure why.
                    final = final.replace(scheme, "", 1).lstrip()
                    tokenbuf = win32security.PySecBufferType(
                        pkg_info["MaxToken"], sspicon.SECBUFFER_TOKEN
                    )
                    tokenbuf.Buffer = base64.b64decode(final.encode("ASCII"))
                    sec_buffer.append(tokenbuf)
                    error, auth = clientauth.authorize(sec_buffer)
                    _logger.debug(
                        "Kerberos Authentication succeeded - error={} authenticated={}".format(
                            error, clientauth.authenticated
                        )
                    )
                except TypeError:
                    pass

            # Regardless of whether or not we finalized our auth context,
            # without a 401 we've got nothing to do. Update the history and return.
            response2.history.append(response)
            return response2

        # Consume content and release the original connection
        # to allow our new request to reuse the same one.
        response2.content
        response2.raw.release_conn()
        request = response2.request.copy()

        # Keep passing the cookies along
        if response2.headers.get("set-cookie"):
            request.headers["Cookie"] = response2.headers.get("set-cookie")

        # Extract challenge message from server
        challenge = [
            val[len(scheme) + 1 :]
            for val in response2.headers.get("WWW-Authenticate", "").split(", ")
            if scheme in val
        ]
        if len(challenge) != 1:
            raise HTTPError(
                "Did not get exactly one {} challenge from server.".format(scheme)
            )

        # Add challenge to security buffer
        tokenbuf = win32security.PySecBufferType(
            pkg_info["MaxToken"], sspicon.SECBUFFER_TOKEN
        )
        tokenbuf.Buffer = base64.b64decode(challenge[0])
        sec_buffer.append(tokenbuf)
        _logger.debug("Got Challenge Token (NTLM)")

        # Perform next authorization step
        try:
            error, auth = clientauth.authorize(sec_buffer)
            request.headers["Authorization"] = "{} {}".format(
                scheme, base64.b64encode(auth[0].Buffer).decode("ASCII")
            )
            _logger.debug(
                "Sending Response - error={} authenticated={}".format(
                    error, clientauth.authenticated
                )
            )
        except pywintypes.error as e:
            _logger.debug("Error calling {}: {}".format(e[1], e[2]), exc_info=e)
            return response

        response3 = response2.connection.send(request, **args)

        if response3.text.lower().find("token required") > -1:
            request4 = self.generate_token(r=response3, scheme=scheme, args=args)
            response4 = response3.connection.send(request4, **args)
            if response4.status_code == 401:
                response4.request.headers.pop("Authorization")
            return self._retry_using_http_Negotiate_auth(
                response=response4, scheme=scheme, args=args
            )

        else:
            # Update the history and return
            response3.history.append(response)
            response3.history.append(response2)
            return response3

    # ----------------------------------------------------------------------
    def _response_hook(self, r, **kwargs):
        """hook logic"""
        if r.status_code == 401:
            for scheme in ("Negotiate", "NTLM"):
                if scheme.lower() in r.headers.get("WWW-Authenticate", "").lower():
                    return self._retry_using_http_Negotiate_auth(r, scheme, kwargs)
        elif r.text.lower().find("token required") > -1:
            for scheme in ("Negotiate", "NTLM"):
                if (
                    scheme.lower()
                    in r.headers.get("WWW-Authenticate", "Negotiate").lower()
                ):
                    request = self.generate_token(r, "NTLM", kwargs)
                    response4 = r.connection.send(request, **kwargs)
                    response4.history.append(r)
                    return response4

    # ----------------------------------------------------------------------
    def __call__(self, r):
        """call method used by requests to register the hook."""
        r.headers["Connection"] = "Keep-Alive"
        r.register_hook("response", self._response_hook)
        return r
